import warnings
from typing import Any, Literal

import matplotlib.patheffects as pe
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnchoredText
from matplotlib.text import Text

from ...utils._parse_units import parse_units
from ...utils.constants import *
from ...utils.ground_sites import get_ground_site
from ...utils.np_array_utils import all_same
from ...utils.overpass import OverpassInfo
from ...utils.read import get_product_infos
from ...utils.read.product.file_info import FileType, ProductDataFrame
from ...utils.time import TimestampLike, format_time_range_text, to_timestamp
from ...utils.typing import HasAxes
from ..color import Color, ColorLike
from .format_strings import wrap_label


def add_text(
    ax: Axes,
    text: str,
    loc: str = "upper right",
    borderpad: float = 0,
    pad: float = 0.4,
    fontsize: str | float | None = None,
    fontweight: str | None = None,
    horizontalalignment: str = "left",
    color: Color | ColorLike | None = "black",
    is_shaded_text: bool = True,
    shade_linewidth: float = 3,
    shade_color: str = "white",
    shade_alpha: float = 0.8,
    is_box: bool = False,
    append_to: AnchoredText | str | None = None,
    zorder: int | float | None = None,
) -> AnchoredText:
    """
    Adds anchored text to a matplotlib Axes with optional shading and styling.

    Args:
        ax (matplotlib.axes.Axes): Target matplotlib Axes.
        text (str): Text string to display.
        loc (str): Anchor location in the Axes (e.g. 'upper right').
        borderpad (float): Padding between text and the border of the box.
        pad (float): Padding between box and the Axes.
        fontsize (str or float, optional): Font size of the text.
        fontweight (str, optional): Font weight (e.g. 'normal', 'bold').
        horizontalalignment (str): Horizontal alignment of the text.
        color (Color or ColorLike, optional): Text color.
        is_shaded_text (bool): If True, apply a white stroke around the text.
        shade_linewidth (float): Width of the stroke line.
        shade_color (str): Color of the stroke.
        shade_alpha (float): Opacity of the stroke.
        is_box (bool): If True, draw a box around the text.
        append_to (AnchoredText or str, optional): Extracts the given text string and adds the new text to it.
        zorder (int | float, optional): Drawing order of the plot element.

    Returns:
        AnchoredText: The text artist added to the Axes.
    """
    old_text: str | None = None
    if isinstance(append_to, AnchoredText):
        old_text = append_to.txt.get_text()
        append_to.remove()
    elif isinstance(append_to, str):
        old_text = append_to

    if isinstance(old_text, str):
        text = f"{old_text}{text}"

    path_effects = None
    if is_shaded_text:
        path_effects = [
            pe.withStroke(
                linewidth=shade_linewidth,
                foreground=shade_color,
                alpha=shade_alpha,
            )
        ]

    text_properties = {
        "size": fontsize,
        "fontweight": fontweight,
        "horizontalalignment": horizontalalignment,
        "path_effects": path_effects,
        "color": color,
    }

    anchored_text = AnchoredText(
        text,
        loc=loc,
        borderpad=borderpad,
        pad=pad,
        prop=text_properties,
        frameon=is_box,
        zorder=zorder,
    )
    ax.add_artist(anchored_text)
    return anchored_text


def add_title(
    ax: Axes,
    title: str,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "center",
    **kwargs,
) -> Text:
    return ax.set_title(title, loc=loc, fontsize=fontsize, **kwargs)


def get_earthcare_frame_string(data: xr.Dataset | ProductDataFrame) -> str:
    text: str = ""

    if isinstance(data, xr.Dataset):
        if "concat_dim" in data.dims:
            if "frame_id" in data and "orbit_number" in data:
                orbit_start = str(data["orbit_number"].values[0]).zfill(5)
                orbit_end = str(data["orbit_number"].values[-1]).zfill(5)
                frame_start = str(data["frame_id"].values[0])
                frame_end = str(data["frame_id"].values[-1])

                if orbit_start == orbit_end:
                    text = (
                        f"{orbit_start}{frame_start}"
                        if frame_start == frame_end
                        else f"{orbit_start}{frame_start}-{frame_end}"
                    )
                else:
                    text = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"
                return text
        elif "orbit_and_frame" in data:
            return str(data["orbit_and_frame"].values)
        elif "frame_id" in data and "orbit_number" in data:
            o = str(data["orbit_number"].values).zfill(5)
            f = str(data["frame_id"].values)
            return f"{o}{f}"

    try:
        df: ProductDataFrame = get_product_infos(data)
    except ValueError as e:
        return text

    if len(df.shape) == 2 and df.shape[0] == 1:
        text = str(df["orbit_and_frame"][0])
    elif len(df.shape) == 2 and df.shape[0] > 1:
        orbit_start = str(df["orbit_number"].iloc[0]).zfill(5)
        orbit_end = str(df["orbit_number"].iloc[-1]).zfill(5)
        frame_start = df["frame_id"].iloc[0]
        frame_end = df["frame_id"].iloc[-1]

        if orbit_start == orbit_end:
            text = (
                f"{orbit_start}{frame_start}"
                if frame_start == frame_end
                else f"{orbit_start}{frame_start}-{frame_end}"
            )
        else:
            text = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"

    return text


def get_earthcare_file_type_baseline_string(data: xr.Dataset | ProductDataFrame) -> str:
    text: str = ""

    if isinstance(data, xr.Dataset):
        if "file_type" in data and "baseline" in data:
            ft = np.atleast_1d(data["file_type"].values)[0]
            ft = FileType.from_input(ft).to_shorthand()
            bl = np.atleast_1d(data["baseline"].values)[0]
            return f"{ft}:{bl}"

    try:
        df: ProductDataFrame = get_product_infos(data)
    except ValueError as e:
        return text

    file_types = df["file_type"]
    baselines = df["baseline"]

    if not all_same(baselines):
        warnings.warn(f"The data contains multiple baselines: {baselines}")

    file_type: str = FileType.from_input(file_types[0]).to_shorthand()
    baseline: str = baselines[0]

    text = f"{file_type}:{baseline}"

    return text


def add_title_earthcare_frame(
    ax: Axes,
    ds: xr.Dataset,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "right",
    color: Color | ColorLike | None = "black",
) -> Text:
    color = Color.from_optional(color)
    text = get_earthcare_frame_string(ds)
    return add_title(ax, text, fontsize=fontsize, loc=loc, color=color)
    # plt.title(text, loc=loc, fontsize=fontsize, color=color)


def add_title_earthcare_time(
    ax: Axes,
    ds: xr.Dataset | None = None,
    time_var: str = TIME_VAR,
    tmin: TimestampLike | None = None,
    tmax: TimestampLike | None = None,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "left",
    color: Color | ColorLike | None = "black",
) -> Text:
    color = Color.from_optional(color)

    _tmin: TimestampLike | None = None
    _tmax: TimestampLike | None = None

    if isinstance(ds, xr.Dataset):
        _tmin = ds[time_var].values[0]
        _tmax = ds[time_var].values[-1]

    if isinstance(tmin, TimestampLike):
        _tmin = to_timestamp(tmin)

    if isinstance(tmax, TimestampLike):
        _tmax = to_timestamp(tmax)

    if isinstance(_tmin, TimestampLike) and isinstance(_tmax, TimestampLike):
        text = format_time_range_text(_tmin, _tmax)
    else:
        raise ValueError(
            f"Missing arguments. At least 'ds' or 'tmin' and 'tmax' must be given."
        )

    return add_title(ax, text, fontsize=fontsize, loc=loc, color=color)


def add_text_overpass_info(
    ax: Axes,
    info: OverpassInfo,
    zorder: int | float | None = 100,
) -> list[AnchoredText]:

    site = get_ground_site(info.site)
    site_name = site.name
    site_altitude = site.altitude
    site_coords = site.coordinates
    radius = info.site_radius_km
    samples = info.samples
    along_track_distance = info.along_track_distance_km
    closest_distance = info.closest_distance_km
    closest_time = info.closest_time

    _site_name = f" ({site_name})" if site_name != "" else ""
    _alt = f" {int(site_altitude)}m" if site_altitude is not None else ""
    _lat = "{:.3f}".format(site_coords[0]) + r"$^\circ\text{N}$"
    _lon = "{:.3f}".format(site_coords[1]) + r"$^\circ\text{E}$"
    _radius = f"Radius: {'{:.0f}'.format(radius)}km"
    _samples = f"Samples: {samples}"
    _along_track_distance = (
        ""
        if along_track_distance is None
        else f"\nAlong-track: {np.round(along_track_distance, decimals=3)}km"
    )
    _closest_distance = (
        ""
        if closest_distance is None
        else f"\nClosest: {np.round(closest_distance, decimals=3)}km at {pd.Timestamp(closest_time).strftime('%H:%M:%S')} UTC"
    )
    info_string = f"{_lat} {_lon}{_alt}{_site_name}\n{_radius}"
    add_text(
        ax,
        info_string,
        loc="upper left",
        horizontalalignment="left",
        fontsize="small",
        zorder=zorder,
    )
    info_string2 = f"{_samples}{_along_track_distance}{_closest_distance}"
    add_text(
        ax,
        info_string2,
        loc="lower left",
        horizontalalignment="left",
        fontsize="small",
        zorder=zorder,
    )
    text = f""

    t1 = add_text(
        ax,
        text,
        zorder=zorder,
    )

    return [t1]


def add_text_product_info(
    ax: Axes,
    ds: xr.Dataset,
    fontsize: str | float = "medium",
    loc: str = "upper right",
    color: Color | ColorLike | None = "black",
    append_to: AnchoredText | str | None = None,
    zorder: int | float | None = 100,
) -> AnchoredText:
    color = Color.from_optional(color)
    text_frame = get_earthcare_frame_string(ds)
    text_type_baseline = get_earthcare_file_type_baseline_string(ds)
    text = f"{text_frame}\n{text_type_baseline}"

    old_text: str | None = None
    if isinstance(append_to, AnchoredText):
        old_text = append_to.txt.get_text()
        append_to.remove()
    elif isinstance(append_to, str):
        old_text = append_to
    if isinstance(old_text, str):
        text = old_text
        if text_frame not in text:
            text = f"{text}\n{text_frame}"
        if text_type_baseline not in text:
            text = f"{text}\n{text_type_baseline}"

    horizontalalignment: str = "center"
    if "left" in loc:
        horizontalalignment = "left"
    elif "right" in loc:
        horizontalalignment = "right"

    return add_text(
        ax=ax,
        text=text,
        fontsize=fontsize,
        loc=loc,
        horizontalalignment=horizontalalignment,
        color=color,
        fontweight="bold",
        zorder=zorder,
    )


def format_var_label(
    name: str | None = None,
    units: str | None = None,
    da: xr.DataArray | None = None,
    label_len: int | None = 40,
) -> str:
    """Format a label with optional units and wrap it to a specified maximum line length.

    Args:
        name (str | None): The base name of the label.
        units (str | None, optional): The units to include in the label. Defaults to None.
        da (xr.DataArray | None, optional): A `xarray.DataArray` from which the label and units
            will be taken, if it has the attributes 'long_name' and 'units'. Defaults to None.
        label_len (int | None, optional): The maximum length of each line. Defaults to 40.

    Returns:
        str: The formatted and wrapped label string.
    """

    if label_len is None:
        label_len = 40

    label: str

    if isinstance(da, xr.DataArray):
        if name is None and hasattr(da, "long_name"):
            name = da.long_name
        if units is None and hasattr(da, "units"):
            units = da.units

    if name is None:
        name = ""
    elif not isinstance(name, str):
        raise TypeError(
            f"Invalid type '{type(name).__name__}' for variable name: {name}. Expected type 'str'."
        )

    label = name

    if isinstance(units, str):
        if units == "":
            pass
        elif units.lower() not in ["-", "none"]:
            label = f"{name} [{parse_units(units, use_latex=True)}]"

    label = wrap_label(label, label_len)

    return label


def add_image_source_label(
    ax: Axes | HasAxes,
    data: (
        Literal["osm", "nasa", "nasagibs", "eumetsat", "mtg", "msg", "esa"] | str | None
    ) = None,
    text: str | None = None,
    loc: str = "lower right",
    fontsize: str = "x-small",
    box_alpha: float = 0.6,
    box_color: str = "white",
    pad: float = 0.2,
    borderpad: float = 0.1,
    change_anchor: bool = False,
    bbox_to_anchor: tuple[float, float] = (1.01, -0.08),
) -> AnchoredText | None:
    """
    Adds a small text label to a plot, intended to display background images sources in map plots.

    Args:
        ax (Axes): The image axes.
        data (Literal[&quot;osm&quot;, &quot;nasa&quot;, &quot;nasagibs&quot;, &quot;eumetsat&quot;, &quot;mtg&quot;, &quot;msg&quot;, &quot;esa&quot;] | None, optional): A tag name used to select a predefiened attribution text. Defaults to None.
        text (str | None, optional): The (manual) attribution text. Defaults to None.
        loc (str, optional): Positioning string for the label in the plot. Defaults to "lower right".
        fontsize (str, optional): Text size. Defaults to "x-small".
        box_alpha (float, optional): Transparency of the label box. Defaults to 0.6.
        box_color (str, optional): Color of the label box. Defaults to "white".
        pad (float, optional): Inside padding between text and box edges. Defaults to 0.2.
        borderpad (float, optional): Outside padding around box. Defaults to 0.1.

    Returns:
        AnchoredText | None: The text object or nothing, if invalid inputs.
    """
    _ax: Axes
    if hasattr(ax, "ax") and isinstance(ax.ax, Axes):
        _ax = ax.ax
    elif isinstance(ax, Axes):
        _ax = ax
    else:
        raise TypeError(f"invalid ax")

    if not isinstance(text, str):
        if not isinstance(data, str):
            return None

        data = str(data).lower()
        data = data.replace(" ", "").replace("-", "").replace("_", "")

        if data == "osm":
            text = "© OSM contributors"
        elif data == "nasa":
            text = "© NASA"
        elif data in ["nasagibs"]:
            text = "© NASA GIBS"
        elif data in ["bluemarble"]:
            text = "Blue Marble © NASA"
        elif data in ["eumetsat"]:
            text = f"© EUMETSAT {pd.Timestamp.now().year}"
        elif data in ["mtg"]:
            text = f"MTG GeoColour\n© EUMETSAT / NASA"
        elif data in ["msg"]:
            text = f"Natural Colour Enhanced RGB\n© EUMETSAT {pd.Timestamp.now().year}"
        elif data == "esa":
            text = "© ESA"
        elif data == "ecmwf":
            text = "© ECMWF"
        else:
            return None

    kwargs: dict[str, Any] = {}
    if change_anchor:
        kwargs["bbox_to_anchor"] = bbox_to_anchor
        kwargs["bbox_transform"] = _ax.transAxes

    at = AnchoredText(
        text,
        loc=loc,
        frameon=True,
        pad=pad,
        borderpad=borderpad,
        prop={"fontsize": fontsize},
        **kwargs,
    )
    at.patch.set_facecolor(box_color)
    at.patch.set_alpha(box_alpha)
    at.patch.set_edgecolor("none")
    _ax.add_artist(at)

    return at



from typing import TYPE_CHECKING, Union, overload

import numpy as np
import xarray as xr

from ...xarray_utils import convert_scalar_var_to_str, merge_datasets


@overload
def read_header_data(source: str) -> xr.Dataset: ...
@overload
def read_header_data(source: xr.Dataset) -> xr.Dataset: ...
def read_header_data(source: str | xr.Dataset) -> xr.Dataset:
    """Opens the product header groups of a EarthCARE file as a `xarray.Dataset`."""
    if isinstance(source, str):
        filepath = source
    elif isinstance(source, xr.Dataset):
        filepath = source.encoding.get("source", None)
        if filepath is None:
            raise ValueError(f"Dataset missing source attribute")
    else:
        raise TypeError("Expected 'str' or 'xarray.Dataset'")

    groups = xr.open_groups(filepath)
    header_groups = {n: g for n, g in groups.items() if "HeaderData" in n}

    # Rename duplicate vars

    all_vars = {}
    header_datasets = []
    for i, (group_name, ds) in enumerate(header_groups.items()):
        ds_new = ds.copy()
        for var in ds.data_vars:
            if var in all_vars:
                new_name = f"{group_name.split('/')[-1]}_{var}"
                ds_new = ds_new.rename({var: new_name})
            else:
                all_vars[var] = True
        header_datasets.append(ds_new)

    ds = xr.merge(header_datasets)

    # Convert timestamps to numpy datetime
    for var in [
        "Creation_Date",
        "Validity_Start",
        "Validity_Stop",
        "ANXTime",
        "frameStartTime",
        "frameStopTime",
        "processingStartTime",
        "processingStopTime",
        "sensingStartTime",
        "sensingStopTime",
        "stateVectorTime",
    ]:
        if var in ds:
            raw = ds[var].values
            formatted = np.char.replace(raw, "UTC=", "")
            ds[var].values = formatted.astype("datetime64[ns]")

    # Ensure that strings are correctly decoded
    for var in ["frameID"]:
        if var in ds:
            ds = convert_scalar_var_to_str(ds, var)

    # Remove dimensions of size == 1
    ds = ds.squeeze()

    return ds


def extract_basic_meta_data_from_header(ds: xr.Dataset) -> xr.Dataset:
    ds["filename"] = ds["File_Name"]
    ds["baseline"] = ds["File_Class"]
    baseline = str(np.str_(ds["File_Class"].values))
    if len(baseline) > 2:
        baseline = baseline[2:]
    ds["baseline"].values = np.array(baseline)

    ds["file_type"] = ds["File_Type"]
    ds["sensing_start_time"] = ds["sensingStartTime"]
    ds["processing_start_time"] = ds["processingStartTime"]
    ds["orbit_number"] = ds["orbitNumber"]
    ds["frame_id"] = ds["frameID"]

    ds["orbit_and_frame"] = ds["frameID"]
    orbit_and_frame = np.str_(ds["orbit_number"].values).zfill(5) + np.str_(
        ds["frame_id"].values
    )
    ds["orbit_and_frame"].values = np.array(orbit_and_frame)

    keep_vars = [
        "filename",
        "file_type",
        "frame_id",
        "orbit_number",
        "orbit_and_frame",
        "baseline",
        "sensing_start_time",
        "processing_start_time",
    ]
    return ds[keep_vars]


def add_header_and_meta_data(
    filepath: str, ds: xr.Dataset, header: bool, meta: bool
) -> xr.Dataset:
    ds_hdr: xr.Dataset | None = None
    if header:
        ds_hdr = read_header_data(filepath)
        ds = merge_datasets(ds_hdr, ds, keep_sec=True)

    if meta:
        if not isinstance(ds_hdr, xr.Dataset):
            ds_hdr = read_header_data(filepath)
        ds_meta = extract_basic_meta_data_from_header(ds_hdr)
        ds = merge_datasets(ds_meta, ds, keep_sec=True)
    return ds



from typing import Literal

import numpy as np
import xarray as xr
from matplotlib.colors import Colormap
from matplotlib.figure import Figure

from ...utils.constants import CM_AS_INCH, FIGURE_HEIGHT_CURTAIN, FIGURE_WIDTH_CURTAIN
from ...utils.read import read_product
from ...utils.time import to_timestamp
from ...utils.typing import ValueRangeLike
from ..color import Color, ColorLike, get_cmap
from ..figure import MapFigure, create_column_figure_layout


def ecswath(
    ds: xr.Dataset | str,
    var: str | None = None,
    n: int = 1,
    time_var: str = "time",
    style: str = "gray",
    border_color: ColorLike | None = "white",
    cmap: str | Colormap | None = None,
    value_range: ValueRangeLike | None = None,
    show_colorbar: bool = True,
    track_color: ColorLike | None = "black",
    color_ticks: ColorLike | None = "white",
    linewidth: float = 3.5,
    linestyle: str = "dashed",
    single_figsize: tuple[float, float] = (3, 8),
    colorbar_position: str | Literal["left", "right", "top", "bottom"] = "bottom",
) -> tuple[Figure, list[MapFigure]]:
    ds = read_product(ds, in_memory=True)

    color_ticks = Color.from_optional(color_ticks) or "white"

    _output = create_column_figure_layout(n, single_figsize=single_figsize)
    fig = _output.fig
    axs = _output.axs
    tmin = to_timestamp(np.nanmin(ds[time_var].values))
    tmax = to_timestamp(np.nanmax(ds[time_var].values))
    tspan = (tmax - tmin) / n
    colorbar_alignment = "center"
    colorbar_length_ratio = "100%"
    if colorbar_position in ["top", "bottom"]:
        if n % 2 == 0:
            colorbar_alignment = "left"
        if n != 1:
            colorbar_length_ratio = "200%"
    map_figs: list[MapFigure] = []
    coastlines_resolution: Literal["10m", "50m", "110m"] = "50m"
    if n > 3:
        coastlines_resolution = "10m"
    for i in range(n):
        _show_colorbar = False
        if colorbar_position in ["top", "bottom"]:
            if i == (n - 1) // 2:
                _show_colorbar = True
        elif i == n - 1:
            _show_colorbar = True
        show_text_time = False
        show_text_frame = False
        if i == 0:
            show_text_time = True
        if i == (n - 1):
            show_text_frame = True
        p = MapFigure(
            ax=axs[i],
            figsize=single_figsize,
            show_right_labels=False,
            show_top_labels=False,
            pad=0,
            style=style,
            show_text_time=show_text_time,
            show_text_frame=show_text_frame,
            border_color=border_color,
            coastlines_resolution=coastlines_resolution,
        )
        p = p.ecplot(
            ds,
            var,
            view="overpass",
            zoom_tmin=tmin + i * tspan,
            zoom_tmax=tmin + (i + 1) * tspan,
            colorbar=show_colorbar & _show_colorbar,
            colorbar_length_ratio=colorbar_length_ratio,
            colorbar_position=colorbar_position,
            colorbar_alignment=colorbar_alignment,
            colorbar_spacing=0.1,
            cmap=None,
            value_range=value_range,
            color=track_color,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        p.grid_lines.ypadding = -3  # type: ignore
        p.grid_lines.ylabel_style = {  # type: ignore
            "color": color_ticks,
            "fontsize": "small",
            "weight": "bold",
        }
        p.grid_lines.xpadding = -3  # type: ignore
        p.grid_lines.xlabel_style = {  # type: ignore
            "color": color_ticks,
            "fontsize": "small",
            "weight": "bold",
        }
        p.grid_lines.zorder = 5  # type: ignore

        map_figs.append(p)
    return fig, map_figs



import numpy as np
import xarray as xr

from ...utils import filter_time
from ...utils.constants import TIME_VAR
from ...utils.read.product import update_rgb_of_mrgr
from ...utils.time import TimeRangeLike


def ensure_updated_msi_rgb_if_required(
    ds: xr.Dataset,
    var: str,
    time_range: TimeRangeLike | None,
    time_var: str = TIME_VAR,
) -> xr.Dataset:
    if var == "rgb" and "swir1" in ds and "nir" in ds and "vis" in ds:
        ds_filtered = filter_time(ds, time_range)
        ds_filtered = update_rgb_of_mrgr(ds_filtered)
        t1 = ds[time_var].values
        t2 = ds_filtered[time_var].values
        mask = np.logical_and(t1 >= t2[0], t1 <= t2[-1])
        ds[var].values[:, mask] = ds_filtered[var].values
        ds[var].values[:, mask] = ds_filtered[var].values
    return ds



import re
from typing import Any, Iterable, List

import numpy as np

from .typing import DistanceRangeLike


def validate_completeness_of_args(
    function_name: str,
    required_names: Iterable[str],
    positional_values: Iterable[Any] | None = None,
    **kwargs,
) -> List[str]:
    """
    Validates that required positional and optional argument groups are complete.

    For example, if `required_names = ['x', 'y']`, this function will:
    - Check if `x` and `y` are both provided in positional arguments (if used).
    - Check if optional arguments like `x1`, `x2`, ..., `y1`, `y2`, ... are all matched
      across the required names (e.g., if `x2` is given, `y2` must also be present).

    Parameters:
        function_name (str): Name of the function (used in error messages).
        required_names (Iterable[str]): Base names of required argument groups (e.g. ['x', 'y']).
        positional_values (Iterable[Any], optional): Positional argument values to validate.
        **kwargs: Optional keyword arguments to validate for matching suffixes.

    Returns:
        List[str]: List of suffixes (as strings) used in optional arguments for the first group.

    Raises:
        TypeError: If any required arguments are missing.
    """
    # Check required positional arguments
    if positional_values is not None:
        missing = [
            required_names[i] for i, v in enumerate(positional_values) if v is None
        ]
        if missing:
            msg = f"{function_name}() missing {len(missing)} required positional argument{'s' if len(missing) > 1 else ''}: "
            msg += ", ".join(f"'{name}'" for name in missing)
            raise TypeError(msg)

    # Check completeness of optional argument groups (e.g., x1/y1, x2/y2, ...)
    suffixes_by_name = [[] for _ in required_names]
    for key in kwargs:
        for i, name in enumerate(required_names):
            match = re.fullmatch(f"{name}(\\d*)", key)
            if match:
                suffix = match.group(1)
                suffixes_by_name[i].append(suffix)

    # Find mismatched suffixes between argument groups
    missing = []
    for i in range(len(suffixes_by_name) - 1):
        for j in range(i + 1, len(suffixes_by_name)):
            suffixes_i = np.array(suffixes_by_name[i])
            suffixes_j = np.array(suffixes_by_name[j])
            missing_from_j = np.setdiff1d(suffixes_i, suffixes_j)
            missing_from_i = np.setdiff1d(suffixes_j, suffixes_i)
            missing += [f"'{required_names[j]}{s}'" for s in missing_from_j]
            missing += [f"'{required_names[i]}{s}'" for s in missing_from_i]

    # Raise error if mismatches were found
    if missing:
        unique_missing = sorted(set(missing))
        msg = f"{function_name}() missing {len(unique_missing)} required argument{'s' if len(unique_missing) > 1 else ''}: "
        msg += ", ".join(unique_missing)
        raise TypeError(msg)

    return suffixes_by_name[0]


def validate_height_range(height_range: DistanceRangeLike) -> tuple[float, float]:
    """Returns validated height range and raises `ValueError` if invalid."""
    if isinstance(height_range, Iterable):
        if len(height_range) == 2:
            if all(
                [
                    isinstance(x, (int, float, np.floating, np.integer))
                    for x in height_range
                ]
            ):
                return float(height_range[0]), float(height_range[1])
    raise ValueError(f"invalid height range: {height_range}")



from matplotlib import colormaps as mpl_cmaps

from .chiljet2 import get_cmap as get_cmap_chiljet2


def get_cmap():
    cmap = get_cmap_chiljet2()
    cmap.name = "radar_reflectivity"
    return cmap



from typing import Any, Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import (
    BoundaryNorm,
    Colormap,
    LinearSegmentedColormap,
    ListedColormap,
    Normalize,
)

from ..color import Color, ColorLike


def copy_extremes(cmap: Union[Colormap, "Cmap"], new_cmap: "Cmap") -> "Cmap":
    new_cmap._rgba_bad = cmap._rgba_bad  # type: ignore
    new_cmap._rgba_over = cmap._rgba_over  # type: ignore
    new_cmap._rgba_under = cmap._rgba_under  # type: ignore
    return new_cmap


def colormap_to_opaque(cmap, N=256):
    if cmap.is_categorical:
        colors = cmap.colors
    else:
        colors = cmap(np.linspace(0, 1, N))
    colors = [Color(c, is_normalized=True) for c in colors]

    colors = [c.set_alpha(1) for c in colors]
    new_cmap = Cmap(
        colors,
        name=cmap.name + "_opaque",
    ).to_categorical(dict(zip(cmap.values, cmap.labels)))
    new_cmap = copy_extremes(cmap, new_cmap)
    return new_cmap


def colormap_to_alphamap(cmap, N=256):
    if cmap.is_categorical:
        colors = cmap.colors
    else:
        colors = cmap(np.linspace(0, 1, N))
    colors = [Color(c, is_normalized=True) for c in colors]

    alpha = [c.alpha for c in colors]
    colors = [Color("black").blend(1 - a) for a in alpha]
    new_cmap = Cmap(colors, name=cmap.name + "_alphamap").to_categorical(
        dict(zip(cmap.values, cmap.labels))
    )
    new_cmap = copy_extremes(cmap, new_cmap)
    return new_cmap


def colormap_to_blended(cmap, N=256):
    if cmap.is_categorical:
        colors = cmap.colors
    else:
        colors = cmap(np.linspace(0, 1, N))
    colors = [Color(c, is_normalized=True) for c in colors]

    colors = [c.blend(c.alpha) for c in colors]
    new_cmap = Cmap(colors, name=cmap.name + "_blended").to_categorical(
        dict(zip(cmap.values, cmap.labels))
    )
    new_cmap = copy_extremes(cmap, new_cmap)
    return new_cmap


class Cmap(ListedColormap):
    """Colormap with categorical, gradient, and circular support.

    This subclass of `matplotlib.colors.ListedColormap` adds utilities for
    continuous and categorical color mappings. Supports labels, ticks,
    normalization, blending, and transparency adjustments.

    Attributes:
        categorical (bool): Whether the colormap is discrete/categorical.
        gradient (bool): Whether the colormap was generated from a gradient.
        circular (bool): Whether the colormap wraps around cyclically.
        ticks (list[float]): Optional tick positions for categorical plots.
        labels (list[str]): Optional labels corresponding to ticks.
        norm (Normalize | None): Normalization strategy for value mapping.
        values (list): Associated values for categorical mapping.
    """

    def __init__(
        self,
        colors,
        name: str = "colormap",
        N: int | None = None,
        categorical: bool = False,
        ticks: List[float] | None = None,
        labels: List[str] | None = None,
        norm: Normalize | None = None,
        values: List | None = None,
        gradient: bool = False,
        circular: bool = False,
    ):
        """Initialize a Cmap.

        Args:
            colors: Sequence of colors (strings or ColorLike objects).
            name (str): Name of the colormap. Defaults to "colormap".
            N (int | None): Number of discrete colors. Defaults to None.
            categorical (bool): Whether the colormap is discrete/categorical. Defaults to False.
            ticks (list[float] | None): Optional tick positions for categorical plots. Defaults to None.
            labels (list[str] | None): Optional labels corresponding to ticks. Defaults to None.
            norm (Normalize | None): Normalization strategy for value mapping. Defaults to None.
            gradient (bool): If True, generate intermediate gradient colors. Defaults to False.
            circular (bool): If True, colormap wraps around cyclically. Defaults to False.
        """
        colors = [Color(c) if isinstance(c, str) else c for c in colors]

        if gradient:
            tmp_cmap = LinearSegmentedColormap.from_list("tmp_cmap", colors, N=256)
            colors = [tmp_cmap(i) for i in range(256)]

        super().__init__(colors, name=name, N=N or len(colors))
        self.categorical = categorical
        self.gradient = gradient
        self.circular = circular
        self.ticks = ticks or []
        self.labels = labels or []
        self.norm = norm
        self.values = values or []

    @classmethod
    def from_colormap(cls, cmap: Colormap, N: int = 256) -> "Cmap":
        """Create a Cmap instance from an existing Matplotlib colormap.

        Args:
            cmap (Colormap): Source colormap to convert.
            N (int): Number of discrete colors (if needed, e.g, for categorical
                colormaps with limited number of colors). Defaults to 256.

        Returns:
            Cmap: New colormap.
        """
        if isinstance(cmap, cls):
            return cmap
        elif isinstance(cmap, ListedColormap):
            colors = list(cmap.colors)  # type: ignore
            if isinstance(colors, np.ndarray) and colors.ndim == 2:
                N = len(colors)
            else:
                N = cmap.N
        else:
            colors = [cmap(x) for x in np.linspace(0, 1, N)]

        _colors = []
        for c in colors:
            if (
                isinstance(c, (np.ndarray, list, tuple))
                and not isinstance(c, str)
                and all([_c <= 1 for _c in c])
            ):
                _colors.append(Color(c, is_normalized=True))  # type: ignore
            else:
                _colors = [Color(c) for c in colors]  # type: ignore
                continue
        new_cmap = cls([c.hex for c in _colors], name=cmap.name, N=N)
        new_cmap = copy_extremes(cmap, new_cmap)
        return new_cmap

    def to_categorical(
        self,
        values_to_labels: Dict[Any, str] | int,
        endpoint: bool | None = None,
        use_discrete: bool | None = None,
    ) -> "Cmap":
        """Convert a colormap to categorical.

        Args:
            values_to_labels (dict | int): Mapping from values to labels, or
                number of categories if int.
            endpoint (bool | None): Whether the last color is included at 1.0.
            use_discrete (bool | None): If True, use the colormap's defined colors directly rather than sampling across its range.

        Returns:
            Cmap: Categorical version of the colormap.
        """
        if isinstance(values_to_labels, int):
            values_to_labels = {i: str(i) for i in range(values_to_labels)}

        values_to_labels = dict(sorted(values_to_labels.items()))

        keys = list(values_to_labels.keys())
        labels = list(values_to_labels.values())
        sorted_values = keys

        n_classes = len(sorted_values)
        bounds = np.array(sorted_values + [sorted_values[-1] + 1]) - 0.5
        norm = BoundaryNorm(bounds, n_classes)

        ticks = [float(t) for t in np.arange(0.5, n_classes)]

        if use_discrete:
            colors = [self(i) for i in range(n_classes)]
        else:
            if not isinstance(endpoint, bool):
                endpoint = not self.circular
            offset = -1 if endpoint else 0
            colors = [self(i / max(n_classes + offset, 1)) for i in range(n_classes)]

        return Cmap(
            colors=colors,
            name=self.name,
            N=n_classes,
            categorical=True,
            gradient=False,
            circular=self.circular,
            ticks=ticks,
            labels=labels,
            norm=norm,
            values=sorted_values,
        )

    def set_alpha(self, value: float) -> "Cmap":
        """Return a copy of the colormap with modified alpha transparency.

        Args:
            value (float): Alpha value in the range [0, 1].

        Returns:
            Cmap: Colormap with updated transparency.
        """
        if not 0 <= value <= 1:
            raise ValueError(
                f"Invalid alpha value: '{value}' (must be in the 0-1 range)"
            )

        new_cmap = Cmap(
            colors=[Color(c).set_alpha(value) for c in np.asarray(self.colors)],
            name=self.name,
            N=self.N,
            categorical=self.categorical,
            gradient=self.gradient,
            circular=self.circular,
            ticks=self.ticks,
            labels=self.labels,
            norm=self.norm,
        )

        if self._rgba_bad is not None:  # type: ignore
            new_cmap._rgba_bad = Color(self._rgba_bad, is_normalized=True).set_alpha(value).rgba  # type: ignore
        if self._rgba_over is not None:  # type: ignore
            new_cmap._rgba_over = Color(self._rgba_over, is_normalized=True).set_alpha(value).rgba  # type: ignore
        if self._rgba_under is not None:  # type: ignore
            new_cmap._rgba_under = Color(self._rgba_under, is_normalized=True).set_alpha(value).rgba  # type: ignore

        return new_cmap

    def blend(self, value: float, blend_color: Color | ColorLike = "white") -> "Cmap":
        """Return a copy of the colormap blended with a second color.

        Args:
            value (float): Blend factor in the range [0, 1].
            blend_color (Color | str): Color to blend with.

        Returns:
            Cmap: Blended colormap.
        """
        if not 0 <= value <= 1:
            raise ValueError(
                f"Invalid blend value: '{value}' (must be in the 0-1 range)"
            )

        new_cmap = Cmap(
            colors=[
                Color(c).blend(value, blend_color) for c in np.asarray(self.colors)
            ],
            name=self.name,
            N=self.N,
            categorical=self.categorical,
            gradient=self.gradient,
            circular=self.circular,
            ticks=self.ticks,
            labels=self.labels,
            norm=self.norm,
        )

        if self._rgba_bad is not None:  # type: ignore
            new_cmap._rgba_bad = Color(self._rgba_bad, is_normalized=True).blend(value, blend_color).rgba  # type: ignore
        if self._rgba_over is not None:  # type: ignore
            new_cmap._rgba_over = Color(self._rgba_over, is_normalized=True).blend(value, blend_color).rgba  # type: ignore
        if self._rgba_under is not None:  # type: ignore
            new_cmap._rgba_under = Color(self._rgba_under, is_normalized=True).blend(value, blend_color).rgba  # type: ignore

        return new_cmap

    @property
    def rgba_list(self) -> list[tuple[float, ...]]:
        """List of RGBA tuples representing all colors in the colormap."""
        return [Color(c, is_normalized=True).rgba for c in np.array(self.colors)]

    # def set_alpha_gradient(self, alpha_input: list) -> "Cmap":
    #     from matplotlib.colors import ListedColormap
    #     from scipy.interpolate import interp1d

    #     # Interpolate to 256 values
    #     n_colors = 256
    #     x_old = np.linspace(0, 1, len(alpha_input))
    #     x_new = np.linspace(0, 1, n_colors)
    #     alpha_interp = interp1d(x_old, alpha_input, kind="linear")(x_new)

    #     # Get base colormap and apply interpolated alpha
    #     colors = self(np.linspace(0, 1, n_colors))
    #     colors[:, 3] = alpha_interp  # Replace alpha channel

    #     # Create transparent colormap
    #     new_cmap = Cmap(colors, name=self.name)

    @property
    def opaque(self) -> "Cmap":
        """Return an opaque version of the colormap (alpha set to 1)."""
        return colormap_to_opaque(self)

    @property
    def alphamap(self) -> "Cmap":
        """Return the alpha-mapped version of the colormap."""
        return colormap_to_alphamap(self)

    @property
    def blended(self) -> "Cmap":
        """Return a blended version of the colormap (predefined blending)."""
        return colormap_to_blended(self)

    def __new__(cls, *args, **kwargs):
        """Allow instantiation from an existing Colormap or standard arguments."""
        if len(args) == 1 and isinstance(args[0], Colormap):
            return cls.from_colormap(args[0])
        return super().__new__(cls)



from logging import Logger
from typing import Literal, Sequence

import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import CM_AS_INCH, DEFAULT_PROFILE_SHOW_STEPS, TIME_VAR
from ....utils.ground_sites import GroundSite
from ....utils.read.product.level1.atl_nom_1b import get_depol_profile
from ....utils.time import TimedeltaLike, TimeRangeLike
from ....utils.typing import DistanceRangeLike
from ....utils.xarray_utils import filter_radius, filter_time
from ...figure import (
    CurtainFigure,
    ECKFigure,
    FigureType,
    MapFigure,
    ProfileFigure,
    create_multi_figure_layout,
)
from .._cli import print_progress
from .._quicklook_results import QuicklookFigure
from ..set_default_height_range import set_none_height_range_to_default


def ecquicklook_anom(
    ds: xr.Dataset,
    vars: list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = True,
    site: GroundSite | str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = (0, 30e3),
    ds_tropopause: xr.Dataset | None = None,
    ds_elevation: xr.Dataset | None = None,
    ds_temperature: xr.Dataset | None = None,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
    mode: Literal["fast", "exact"] = "fast",
) -> QuicklookFigure:
    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    height_range = set_none_height_range_to_default(height_range, 0, 30e3)

    if ds_elevation is None:
        ds_elevation = ds

    if vars is None:
        vars = [
            "mie_attenuated_backscatter",
            "rayleigh_attenuated_backscatter",
            "crosspolar_attenuated_backscatter",
            "depol_ratio",
        ]

    wspaces: list[float] = []

    if show_maps:
        map_rows = [FigureType.MAP_1_ROW, FigureType.MAP_1_ROW]
        wspaces.append(3.5 * CM_AS_INCH)
    else:
        map_rows = None

    main_rows = [FigureType.CURTAIN for _ in vars]

    if show_zoom:
        zoom_rows = [FigureType.CURTAIN for _ in vars]
        wspaces.append(4 * CM_AS_INCH)
    else:
        zoom_rows = None

    if show_profile:
        profile_rows = [FigureType.PROFILE for _ in vars]
        if show_zoom:
            wspaces.append(2.5 * CM_AS_INCH)
        else:
            wspaces.append(4 * CM_AS_INCH)
    else:
        profile_rows = None

    if logger:
        print_progress(f"layout", log_msg_prefix=log_msg_prefix, logger=logger)

    output = create_multi_figure_layout(
        map_rows=map_rows,
        rows=main_rows,
        zoom_rows=zoom_rows,
        profile_rows=profile_rows,
        wspace=wspaces,
    )

    fig = output.fig
    axs_map = output.axs_map
    axs_main = output.axs
    axs_zoom = output.axs_zoom
    axs_profile = output.axs_profile

    map_figs: list[ECKFigure] = []
    main_figs: list[ECKFigure] = []
    zoom_figs: list[ECKFigure] = []
    profile_figs: list[ECKFigure] = []

    if show_maps:
        if logger:
            print_progress(f"map globe", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(ax=axs_map[0])
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

        if logger:
            print_progress(f"map zoomed", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(
            ax=axs_map[1],
            style="blue_marble",
            coastlines_resolution="50m",
            show_night_shade=False,
            show_right_labels=False,
            show_top_labels=False,
        )
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            view="overpass",
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

    for i, var in enumerate(vars):
        if logger:
            print_progress(
                f"curtain: {var=}", log_msg_prefix=log_msg_prefix, logger=logger
            )
        cf = CurtainFigure(ax=axs_main[i], mode=mode)
        cf = cf.ecplot(
            ds,
            var,
            site=site,
            radius_km=radius_km,
            selection_time_range=time_range,
            height_range=height_range,
            selection_max_time_margin=selection_max_time_margin,
        )
        if ds_tropopause:
            if logger:
                print_progress(
                    f"tropopause", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_tropopause(ds_tropopause)
        if ds_elevation:
            if logger:
                print_progress(
                    f"elevation", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_elevation(ds_elevation)
        if ds_temperature:
            if logger:
                print_progress(
                    f"temperature", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_temperature(ds_temperature)

        main_figs.append(cf)

    if show_zoom or show_profile:
        _ds = ds.copy()
        if site:
            _ds = filter_radius(_ds, radius_km=radius_km, site=site)
        elif time_range:
            _ds = filter_time(_ds, time_range)

        if show_zoom:
            _time_range = _ds[TIME_VAR].values[[0, -1]]
            for i, var in enumerate(vars):
                if logger:
                    print_progress(
                        f"{log_msg_prefix}curtain zoomed: {var=} ...",
                        log_msg_prefix=log_msg_prefix,
                        logger=logger,
                    )
                cf = CurtainFigure(
                    ax=axs_zoom[i],
                    num_ticks=4,
                    show_height_left=False,
                    show_height_right=True,
                    mode=mode,
                )
                cf = cf.ecplot(
                    ds,
                    var,
                    colorbar=False,
                    time_range=_time_range,
                    height_range=height_range,
                )
                if ds_tropopause:
                    if logger:
                        print_progress(
                            f"tropopause zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_tropopause(ds_tropopause)
                if ds_elevation:
                    if logger:
                        print_progress(
                            f"elevation zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_elevation(ds_elevation)
                if ds_temperature:
                    if logger:
                        print_progress(
                            f"temperature zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_temperature(ds_temperature)

                zoom_figs.append(cf)

        if show_profile:
            for i, var in enumerate(vars):
                if logger:
                    print_progress(
                        f"profile: {var=}", log_msg_prefix=log_msg_prefix, logger=logger
                    )
                _var = "depol_ratio_from_means"
                if var == "depol_ratio":
                    _ds[_var].values[:] = get_depol_profile(_ds)

                pf = ProfileFigure(
                    ax=axs_profile[i],
                    flip_height_axis=True,
                )
                pf = pf.ecplot(
                    _ds,
                    _var,
                    height_range=height_range,
                    color="ec:red",
                    show_steps=show_steps,
                )

                profile_figs.append(pf)

    subfigs: list[list[ECKFigure]] = [map_figs, main_figs, zoom_figs, profile_figs]

    _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
    if logger:
        print_progress(
            f"Plot created (time taken {_dtime}).",
            is_last=True,
            log_msg_prefix=log_msg_prefix,
            logger=logger,
        )

    return QuicklookFigure(fig, subfigs)



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap():
    cmap_data = [
        ("#c5c9c7", -1, "Unknown"),
        ("#a2653e", 0, "Surface"),
        ("#ffffff", 1, "Clear"),
        ("#ff474c", 2, "Rain in clutter"),
        ("#0504aa", 3, "Snow in clutter"),
        ("#009337", 4, "Cloud in clutter"),
        ("#840000", 5, "Heavy rain"),
        ("#042e60", 6, "Heavy mixed-phase precip."),
        ("#d8dcd6", 7, "Clear (possible liquid)"),
        ("#ffff84", 8, "Liquid cloud"),
        ("#f5bf03", 9, "Drizzling liquid cloud"),
        ("#f97306", 10, "Warm rain"),
        ("#ff000d", 11, "Cold rain"),
        ("#5539cc", 12, "Melting snow"),
        ("#2976bb", 13, "Snow (possible liquid)"),
        ("#0d75f8", 14, "Snow"),
        ("#014182", 15, "Rimed snow (possible liquid)"),
        ("#017b92", 16, "Rimed snow + s'cooled liquid"),
        ("#06b48b", 17, "Snow + liquid"),
        ("#aaff32", 18, "S'cooled liquid cloud"),
        ("#6dedfd", 19, "Ice cloud (possible liquid)"),
        ("#01f9c6", 20, "Ice + liquid cloud"),
        ("#7bc8f6", 21, "Ice cloud"),
        ("#d7fffe", 22, "Strat. ice"),
        ("#a2cffe", 23, "STS (PSC Type I)"),
        ("#04d9ff", 24, "NAT (PSC Type II)"),
        ("#7a9703", 25, "Insects"),
        ("#b2996e", 26, "Dust"),
        ("#ffbacd", 27, "Sea salt"),
        ("#d99b82", 28, "Continental pollution"),
        ("#947e94", 29, "Smoke"),
        ("#856798", 30, "Dusty smoke"),
        ("#ac86a8", 31, "Dusty mix"),
        ("#59656d", 32, "Strat. ash"),
        ("#76424e", 33, "Strat. sulfate"),
        ("#363737", 34, "Strat. smoke"),
    ]
    colors = [c for c, _, _ in cmap_data]
    definitions = {k: l for _, k, l in cmap_data}
    cmap = Cmap(colors=colors, name="synergetic_tc").to_categorical(definitions)
    return cmap



from dataclasses import dataclass
from logging import Logger

from ._eo_collection import EOCollection
from ._request import get_request_json


@dataclass
class STACQueryParameter:
    name: str
    enum: list[str] | None


def _get_stac_base_parameters() -> list[STACQueryParameter]:
    """Return list of standard STAC API parameters.

    see: https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/item-search#query-parameter-table
    """
    return [
        STACQueryParameter(
            name="limit",
            enum=None,
        ),
        STACQueryParameter(
            name="datetime",
            enum=None,
        ),
        STACQueryParameter(
            name="bbox",
            enum=None,
        ),
        # TODO: Fix handling of these additional non standard parameters (e.g., fails for X-MET)
        STACQueryParameter(
            name="radius",
            enum=None,
        ),
        STACQueryParameter(
            name="lat",
            enum=None,
        ),
        STACQueryParameter(
            name="lon",
            enum=None,
        ),
    ]


def get_available_parameters(
    collection: EOCollection,
    logger: Logger | None = None,
) -> list[STACQueryParameter]:
    """Requests queryables (STAC API parameters) of a selected collection and returns available ones."""
    url_queryables = collection.url_queryables
    if not isinstance(url_queryables, str):
        return []

    data = get_request_json(url=url_queryables, logger=logger)

    raw_paraemters: dict | None = data.get("properties")
    if not isinstance(raw_paraemters, dict):
        raise ValueError(
            f'missing "properties" field in response of "{url_queryables}"'
        )

    eo_parameters: list[STACQueryParameter] = _get_stac_base_parameters()

    for name, info in raw_paraemters.items():
        eop = STACQueryParameter(
            name=name,
            enum=info.get("enum", None),
        )
        eo_parameters.append(eop)

    return eo_parameters



from typing import Sequence

import numpy as np
from numpy.typing import NDArray


def select_value_range(
    data: Sequence | NDArray,
    value_range: Sequence | NDArray | None,
    pad_frac: float = 0.0,
    use_min_max: bool = False,
) -> tuple[float, float]:
    data = np.asarray(data)

    vmin = np.nan
    vmax = np.nan

    if isinstance(value_range, (Sequence, np.ndarray)):
        if len(value_range) > 1:
            _vmin = value_range[0]
            _vmax = value_range[-1]
        if isinstance(_vmin, (int, float)):
            vmin = float(_vmin)
        if isinstance(_vmax, (int, float)):
            vmax = float(_vmax)

    if np.isnan(vmin):
        if use_min_max:
            vmin = np.nanmin(data)
        else:
            vmin = np.nanpercentile(data, 1)
    if np.isnan(vmax):
        if use_min_max:
            vmax = np.nanmax(data)
        else:
            vmax = np.nanpercentile(data, 99)

    vrange = vmax - vmin

    new_value_range = (vmin - vrange * pad_frac, vmax + vrange * pad_frac)

    return new_value_range



import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray


def _get_sec_of_day(time: NDArray):
    sod = (time - time.astype("datetime64[D]")) / np.timedelta64(1, "s")
    return sod


def _get_hour_of_day(time: NDArray):
    hod = _get_sec_of_day(time) / 3600.0
    return hod


def _get_day_of_year(time: NDArray):
    year_start = time.astype("datetime64[Y]")
    doy = (time.astype("datetime64[D]") - year_start) / np.timedelta64(1, "D")
    doy = doy.astype(float) + 1.0
    return doy


def get_day_night_mask(
    time: pd.DatetimeIndex | ArrayLike,
    lats: ArrayLike,
    lons: ArrayLike,
    sun_altitude_threshold: float = 0.0,
) -> NDArray[np.bool]:
    """
    Calculates a day/night mask from UTC timestamps and lat/lon positions.

    Args:
        times (pandas.DatetimeIndex | NDArray):
            UTC Timestamps.
        lats (float | NDArray):
            Latitude(s) in degrees (scalar or same length as `times`).
        lons (float | NDArray):
            Longitude(s) in degrees, positive east (scalar or same length as `times`).
        sun_altitude_threshold (float, optional):
            Threshold in degrees for "day" (0 = horizon, -6 = civil twilight,
            -12 = nautical twilight, -18 = astronomical twilight)

    Returns:
        NDArray[bool]: Day/night mask (True = day, False = night)

    References:
        NOAA Global Monitoring Laboratory, "Low Accuracy Equations"
            https://gml.noaa.gov/grad/solcalc/sollinks.html (accessed 2025-01-21)
        Wikipedia contributors (2025), "Hour angle"
            https://en.wikipedia.org/wiki/Hour_angle?utm_source=chatgpt.com (accessed 2025-01-21)
    """

    if isinstance(time, pd.DatetimeIndex):
        time_np = time.values.astype("datetime64[s]")
    else:
        time_np = np.asarray(time).astype("datetime64[s]")

    hour = _get_hour_of_day(time_np)

    day_of_year = _get_day_of_year(time_np)

    lats = np.asarray(lats, dtype=float)
    lons = np.asarray(lons, dtype=float)
    n = len(hour)
    if lats.ndim == 0:
        lats = np.full(n, lats)
    if lons.ndim == 0:
        lons = np.full(n, lons)

    # Fractional year
    gamma_rad = 2.0 * np.pi / 365.0 * (day_of_year - 1.0 + (hour - 12.0) / 24.0)

    # Solar declination angle
    decl_rad = (
        0.006918
        - 0.399912 * np.cos(gamma_rad)
        + 0.070257 * np.sin(gamma_rad)
        - 0.006758 * np.cos(2 * gamma_rad)
        + 0.000907 * np.sin(2 * gamma_rad)
        - 0.002697 * np.cos(3 * gamma_rad)
        + 0.00148 * np.sin(3 * gamma_rad)
    )

    # Local solar time
    lst = (hour + lons / 15.0) % 24.0

    # Solar hour angle (0 deg at solar noon, 15 deg per hour, negative=morning, positive=afternoon)
    hour_angle = np.deg2rad(15.0 * (lst - 12.0))

    # Solar elevation angle
    lat_rad = np.deg2rad(lats)
    sin_elev = np.sin(lat_rad) * np.sin(decl_rad) + np.cos(lat_rad) * np.cos(
        decl_rad
    ) * np.cos(hour_angle)
    elev_deg = np.rad2deg(np.arcsin(np.clip(sin_elev, -1.0, 1.0)))

    day_night_mask = elev_deg > sun_altitude_threshold
    return day_night_mask



import numpy as np


def get_decimal_shift(x: float) -> int:
    """Returns the base-10 exponent needed to scale `x` into the range [1.0, 10.0).

    For example:
        - x = 0.00034 -> exponent = 4  -> x * 10**4  = 3.4
        - x = 0.564   -> exponent = 1  -> x * 10**1  = 5.64
        - x = 1234.5  -> exponent = -3 -> x * 10**-3 = 1.2345
    """
    if x == 0:
        return 0
    return -int(np.floor(np.log10(abs(x))))



from typing import Iterable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray

from ..constants import ALONG_TRACK_DIM, TIME_VAR
from ..np_array_utils import pad_true_sequence
from ..time import TimeRangeLike, TimestampLike, to_timestamp


def get_time_range(
    ds: xr.Dataset,
    time_range: TimeRangeLike | Iterable | None,
    time_var: str = TIME_VAR,
) -> tuple[pd.Timestamp, pd.Timestamp]:
    """
    Ensures a complete time range by filling in missing start or end values using dataset boundaries.

    Args:
        ds (xr.Dataset): Dataset containing the time variable.
        time_var (str): Name of the time variable in the dataset.
        time_range (TimeRangeLike | Iterable | None): A two-element list, tuple or array containing start and end times,
            which may be strings, pandas Timestamps, or None.

    Returns:
        List[pd.Timestamp]: A complete [start, end] time range as pandas Timestamps.
    """
    if isinstance(time_range, (Sequence, np.ndarray)) and not isinstance(
        time_range, str
    ):
        if len(time_range) >= 2:
            time_range = [
                time_range[0],
                time_range[-1],
            ]
        else:
            raise ValueError(f"invalid time range '{time_range}'")
    elif time_range is None:
        time_range = [None, None]
    else:
        raise TypeError(
            f"Invalid type '{type(time_range).__name__}' for time_range. Expected a 2-element sequence (tuple or list)."
        )

    new_time_range: list[pd.Timestamp] = [pd.Timestamp(0), pd.Timestamp(0)]
    if time_range[0] is None:
        new_time_range[0] = to_timestamp(ds[time_var].values[0])
    else:
        new_time_range[0] = to_timestamp(time_range[0])

    if time_range[1] is None:
        new_time_range[1] = to_timestamp(ds[time_var].values[-1])
    else:
        new_time_range[1] = to_timestamp(time_range[1])

    return (new_time_range[0], new_time_range[1])


def get_filter_time_mask(
    ds: xr.Dataset,
    time_range: TimeRangeLike | Iterable | None = None,
    timestamp: TimestampLike | None = None,
    only_center: bool = False,
    time_var: str = TIME_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    pad_idxs: int = 0,
) -> NDArray:
    times = ds[time_var].values
    mask: NDArray[np.bool_] = np.full(times.shape, False, dtype=bool)
    if timestamp is not None:
        timestamp = to_timestamp(timestamp)

        tmin = to_timestamp(times[0])
        tmax = to_timestamp(times[-1])

        if not tmin <= timestamp <= tmax:
            raise ValueError(
                f"Timestamp {timestamp} lies outside of the dataset's time range ({tmin} -> {tmax})"
            )

        idx = np.argmin(np.abs(times - timestamp.to_numpy()))
        mask[idx] = True
    else:
        time_range = get_time_range(ds, time_range=time_range, time_var=time_var)

        mask = (times >= np.min([time_range[0], time_range[1]])) & (
            times <= np.max([time_range[0], time_range[1]])
        )

    if only_center:
        mask_true_idxs = np.where(mask)[0]
        if len(mask_true_idxs) > 0:
            idx_center = mask_true_idxs[len(mask_true_idxs) // 2]
            mask[:] = False
            mask[idx_center] = True

    mask = pad_true_sequence(mask, pad_idxs)
    return mask


def filter_time(
    ds: xr.Dataset,
    time_range: TimeRangeLike | Iterable | None = None,
    timestamp: TimestampLike | None = None,
    only_center: bool = False,
    time_var: str = TIME_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    pad_idxs: int = 0,
) -> xr.Dataset:
    """
    Filters an xarray Dataset to include only samples within a given time range.

    Args:
        ds (xr.Dataset): The input dataset containing a time coordinate.
        time_range (TimeRangeLike | Iterable | None):
            Start and end time of the range to filter, as strings or pandas timestamps. Defaults to None.
        timestamp (TimestampLike | None): A single timestamp for which the closest sample to return. Defaults to None.
        only_center (bool, optional): If True, only the sample at the center index of selection is returned. Defaults to False.
        time_var (str, optional): Name of the time variable in `ds`. Defaults to TIME_VAR.
        along_track_dim (str, optional): Dimension name along which time is defined. Defaults to ALONG_TRACK_DIM.
        pad_idxs (int, optional): Number of additional samples added at both sides of the selection. Defaults to 0.

    Returns:
        xr.Dataset: Subset of `ds` containing only samples within the specified time range.
    """
    if time_range is not None and timestamp is not None:
        raise ValueError(
            f"Can not use both arguments time_range and timestamp at the same time."
        )

    mask = get_filter_time_mask(
        ds=ds,
        time_range=time_range,
        timestamp=timestamp,
        only_center=only_center,
        time_var=time_var,
        along_track_dim=along_track_dim,
        pad_idxs=pad_idxs,
    )

    if np.sum(mask) == 0:
        times = ds[time_var].values
        msg = (
            f"No data falls into the given time range!\n"
            f"In the dataset time ranges from {times[0]} to {times[-1]}.\n"
        )
        raise ValueError(msg)

    da_mask: xr.DataArray = xr.DataArray(mask, dims=[along_track_dim], name=time_var)

    ds_new: xr.Dataset = xr.Dataset(
        {
            var: (
                ds[var].copy().where(da_mask, drop=True)
                if along_track_dim in ds[var].dims
                else ds[var].copy()
            )
            for var in ds.data_vars
        }
    )
    ds_new.attrs = ds.attrs.copy()
    ds_new.encoding = ds.encoding.copy()

    return ds_new



from matplotlib.colors import Colormap

from .cmap import Cmap
from .colormap import _get_custom_cmaps, get_cmap, rename_cmap
from .shift import shift_cmap

cmaps: dict[str, Colormap] = _get_custom_cmaps()
"""List of custom colormaps for earthcarekit."""



import re
from typing import Final

_LATEST: Final[str] = "latest"


def parse_product_version(product_version: str | None) -> str:
    if product_version is None:
        return _LATEST

    product_version_pattern = "[a-zA-Z]{2}"
    match = re.fullmatch(product_version_pattern, product_version)
    if match:
        return product_version.upper()
    elif product_version.lower() == _LATEST.lower():
        return _LATEST
    else:
        raise ValueError(
            f'bad product_version string "{product_version}". expected 2 alphabetical letters (A-Z) or "latest"'
        )



from logging import Logger

from ._parse_product_type import parse_product_type_and_version
from ._parse_product_version import parse_product_version
from ._types import ProductTypeVersion


def parse_products(
    product_type: list[str],
    product_version: str | None,
    logger: Logger | None = None,
) -> list[ProductTypeVersion]:
    v = parse_product_version(product_version)
    return [
        parse_product_type_and_version(p, default_version=v, logger=logger)
        for p in product_type
    ]



from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

from ..plot.figure import ProfileFigure
from ..utils.profile_data import ProfileData
from ..utils.read.product.level1.atl_nom_1b import add_depol_ratio
from ..utils.typing import DistanceRangeLike, validate_numeric_range


@dataclass(frozen=True)
class _ANOMDepolCalculationResults:
    ds: xr.Dataset
    selection_height_range: tuple[float, float]
    dpol_mean: float
    dpol_std: float
    cpol_mean: float
    cpol_std_t: float
    cpol_std_z: float
    cpol_std: float
    xpol_mean: float
    xpol_std_t: float
    xpol_std_z: float
    xpol_std: float
    error: float
    error_t: float
    error_z: float

    @property
    def stats(self) -> pd.DataFrame:
        return pd.DataFrame(
            dict(
                hmin=[self.selection_height_range[0]],
                hmax=[self.selection_height_range[1]],
                dpol_mean=[self.dpol_mean],
                dpol_std=[self.dpol_std],
                error_t=[self.error_t],
                error_z=[self.error_z],
                error=[self.error],
                cpol_mean=[self.cpol_mean],
                cpol_std_t=[self.cpol_std_t],
                cpol_std_z=[self.cpol_std_z],
                cpol_std=[self.cpol_std],
                xpol_mean=[self.xpol_mean],
                xpol_std_t=[self.xpol_std_t],
                xpol_std_z=[self.xpol_std_z],
                xpol_std=[self.xpol_std],
            )
        )

    def print(self) -> None:
        print(f"===========================================")
        print(f" Statistics: A-NOM depol ratio calculation")
        print(f"===========================================")
        print(
            f" Height range = {self.selection_height_range[0]} to {self.selection_height_range[1]} meters"
        )
        print(f" mean(depol)  = {self.dpol_mean:.4f}")
        print(f" std(depol)   = {self.dpol_std:.4f}")
        print(f"-------------------------------------------")
        print(f" error_t      = {self.error_t:.4f}")
        print(f" error_z      = {self.error_z:.4f}")
        print(f" error        = {self.error:.4f}")
        print(f"-------------------------------------------")
        print(f" mean(cpol)   = {self.cpol_mean:.4e}")
        print(f" std_t(cpol)  = {self.cpol_std_t:.4e}")
        print(f" std_z(cpol)  = {self.cpol_std_z:.4e}")
        print(f" std(cpol)    = {self.cpol_std:.4e}")
        print(f"-------------------------------------------")
        print(f" mean(xpol)   = {self.xpol_mean:.4e}")
        print(f" std_t(xpol)  = {self.xpol_std_t:.4e}")
        print(f" std_z(xpol)  = {self.xpol_std_z:.4e}")
        print(f" std(xpol)    = {self.xpol_std:.4e}")
        print(f"===========================================")

    def plot(
        self,
        profile_figure_kwargs: dict[str, Any] = dict(),
        **kwargs,
    ) -> ProfileFigure:
        if "value_range" not in kwargs:
            kwargs["value_range"] = (0, 0.6)
        return ProfileFigure(**profile_figure_kwargs).ecplot(
            ds=self.ds,
            var="depol_ratio_from_means",
            selection_height_range=self.selection_height_range,
            **kwargs,
        )


def perform_anom_depol_statistics(
    ds_anom: xr.Dataset,
    selection_height_range: DistanceRangeLike,
    **kwargs,
) -> _ANOMDepolCalculationResults:
    """
    Calculate depolarization statistics and uncertainties within a height range.

    This function adds the depol. ratio (`DPOL`) calculated from co- (`CPOL`) and cross-polarized (`XPOL`)
    attenuated backscatter to the dataset (ATL_NOM_1B) and computes related statistics.
    Mean values and standard deviations are calculated for `CPOL`, `XPOL`, and `DPOL` within the selected
    height range. Variability is separated into vertical and temporal components. Errors of `DPOL` are derived
    using error propagation for the `XPOL`/`CPOL` ratio.

    Args:
        ds_anom (xr.Dataset): ATL_NOM_1B dataset with cross- and co-polar signals.
        selection_height_range (DistanceRangeLike): Height range for statistics.

    Returns:
        _ANOMDepolCalculationResults: Results container with

            - Mean and standard deviation of depolarization ratio.
            - Mean, vertical, temporal, and combined spreads for co- and cross-polar signals.
            - Propagated uncertainty of δ (total, vertical, temporal).
            - Input dataset with depolarization ratio added.

    Example:
        ```python
        import earthcarekit as eck

        ft = "ANOM"
        oaf = "01508B"
        site = "dushanbe"
        radius_km = 100
        sel_hrange = (1e3, 4e3)

        # # Optionally, download required data
        # eck.ecdownload(file_type=ft, orbit_and_frame=oaf)

        df = eck.search_product(file_type=ft, orbit_and_frame=oaf)
        fp = df.filepath[-1]

        with eck.read_any(fp) as ds:
            ds = eck.filter_radius(ds, radius_km=radius_km, site=site)
            results = eck.perform_anom_depol_statistics(ds, sel_hrange)
            results.print()  # prints statistics

            # # Optionally, save statistics as CSV file
            # results.stats.to_csv("./stats.csv")

            # # Optionally, save profile figure as PNG file
            # fig = results.plot(height_range=(0, 10e3))
            # eck.save_plot(fig, filepath="./depol_profile.png")
        ```
    """

    selection_height_range = validate_numeric_range(selection_height_range)

    ds_anom = add_depol_ratio(ds_anom, **kwargs)

    cpol_p: ProfileData = ProfileData.from_dataset(
        ds_anom, var="cpol_cleaned_for_depol_calculation"
    )
    xpol_p: ProfileData = ProfileData.from_dataset(
        ds_anom, var="xpol_cleaned_for_depol_calculation"
    )
    cpol_mean_p: ProfileData = cpol_p.mean()
    xpol_mean_p: ProfileData = xpol_p.mean()
    dpol_mean_p: ProfileData = xpol_mean_p / cpol_mean_p
    cpol_std_p: ProfileData = cpol_p.std()
    xpol_std_p: ProfileData = xpol_p.std()

    cpol_stats = cpol_p.stats(selection_height_range)
    xpol_stats = xpol_p.stats(selection_height_range)
    dpol_stats = dpol_mean_p.stats(selection_height_range)

    dpol_mean: float = dpol_stats.mean
    dpol_std: float = dpol_stats.std
    cpol_mean: float = cpol_stats.mean
    xpol_mean: float = xpol_stats.mean
    cpol_std_t: float = cpol_std_p.stats(selection_height_range).std
    xpol_std_t: float = xpol_std_p.stats(selection_height_range).std
    cpol_std_z: float = cpol_stats.std
    xpol_std_z: float = xpol_stats.std
    cpol_std: float = cpol_std_t + cpol_std_z
    xpol_std: float = xpol_std_t + xpol_std_z

    calc_error = lambda xsd, csd: np.sqrt(
        (xsd / cpol_mean) ** 2 + (((xpol_mean / (cpol_mean**2)) * csd) ** 2)
    )
    error = calc_error(xpol_std, cpol_std)
    error_z = calc_error(xpol_std_z, cpol_std_z)
    error_t = calc_error(xpol_std_t, cpol_std_t)

    return _ANOMDepolCalculationResults(
        ds=ds_anom.copy(),
        selection_height_range=selection_height_range,
        dpol_mean=dpol_mean,
        dpol_std=dpol_std,
        cpol_mean=cpol_mean,
        cpol_std_t=cpol_std_t,
        cpol_std_z=cpol_std_z,
        cpol_std=cpol_std,
        xpol_mean=xpol_mean,
        xpol_std_t=xpol_std_t,
        xpol_std_z=xpol_std_z,
        xpol_std=xpol_std,
        error=error,
        error_t=error_t,
        error_z=error_z,
    )



from logging import Logger
from typing import Sequence

from ._parse_all_orbit_and_frame_inputs import parse_all_orbit_and_frame_inputs
from ._parse_dirpaths import parse_path_to_data
from ._parse_geo_search_bbox import parse_bbox_search
from ._parse_geo_search_radius import parse_radius_search
from ._parse_path_to_config import parse_path_to_config
from ._parse_product_type import parse_product_type_and_version
from ._parse_products import parse_products
from ._parse_time import parse_time
from ._types import (
    FrameIDStr,
    OrbitFrameStr,
    OrbitInt,
    ProductTypeVersion,
    TimestampStr,
    _BBoxSearch,
    _OrbitFrameInputs,
    _RadiusSearch,
    _SearchInputs,
    _TimestampInputs,
)


def parse_search_inputs(
    product_type: Sequence[str],
    baseline: str | None = None,
    orbit_number: Sequence[OrbitInt] | None = None,
    start_orbit_number: OrbitInt | None = None,
    end_orbit_number: OrbitInt | None = None,
    frame_id: Sequence[FrameIDStr] | None = None,
    orbit_and_frame: Sequence[OrbitFrameStr] | None = None,
    start_orbit_and_frame: OrbitFrameStr | None = None,
    end_orbit_and_frame: OrbitFrameStr | None = None,
    timestamps: Sequence[TimestampStr] | None = None,
    start_time: TimestampStr | None = None,
    end_time: TimestampStr | None = None,
    radius_search: Sequence[str | float] | None = None,
    bounding_box: Sequence[str | float] | None = None,
    logger: Logger | None = None,
) -> _SearchInputs:

    def _to_list(seq: Sequence | None) -> list | None:
        return None if seq is None else list(seq)

    product_inputs: list[ProductTypeVersion] = parse_products(
        [pt for pt in product_type],
        baseline,
        logger=logger,
    )

    orbit_frame_inputs: _OrbitFrameInputs = parse_all_orbit_and_frame_inputs(
        args_orbit_number=_to_list(orbit_number),
        args_start_orbit_number=start_orbit_number,
        args_end_orbit_number=end_orbit_number,
        args_frame_id=_to_list(frame_id),
        args_orbit_and_frame=_to_list(orbit_and_frame),
        args_start_orbit_and_frame=start_orbit_and_frame,
        args_end_orbit_and_frame=end_orbit_and_frame,
        logger=logger,
    )

    time_inputs: _TimestampInputs = parse_time(
        _to_list(timestamps),
        start_time,
        end_time,
        logger=logger,
    )

    radius_search_inputs: _RadiusSearch = parse_radius_search(_to_list(radius_search))
    bbox_search_inputs: _BBoxSearch = parse_bbox_search(_to_list(bounding_box))

    return _SearchInputs(
        products=product_inputs,
        orbit_and_frames=orbit_frame_inputs,
        timestamps=time_inputs,
        radius_search=radius_search_inputs,
        bbox_search=bbox_search_inputs,
    )



import os
import re
import time

import pandas as pd


def remove_old_logs(
    max_num_logs: int | None = None,
    max_age_logs: pd.Timedelta | None = None,
    log_file_prefix: str = "",
) -> None:
    """Deletes old log files depending on given maximum file number and/or age"""
    logs_dirpath = os.path.abspath("logs")

    if os.path.exists(logs_dirpath):
        pattern = r".*" + log_file_prefix + r"[0-9]{8}T[0-9]{6}(|_[0-9]*).log"
        if max_num_logs:
            old_logs = [
                os.path.abspath(os.path.join(logs_dirpath, fp))
                for fp in os.listdir(logs_dirpath)
                if re.search(pattern, fp)
            ]
            if len(old_logs) > max_num_logs - 1:
                old_logs.sort(reverse=True)
                for log in old_logs[max_num_logs - 1 : :]:
                    os.remove(log)

        if max_age_logs:
            current_time = pd.Timestamp(
                time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(time.time()))
            )
            last_allowed_time = current_time - max_age_logs
            old_logs = [
                os.path.abspath(os.path.join(logs_dirpath, fp))
                for fp in os.listdir(logs_dirpath)
                if re.search(pattern, fp)
            ]
            for log in old_logs:
                log_time = pd.Timestamp(
                    os.path.basename(log).split(".")[0].split("_")[-1]
                )
                if log_time < last_allowed_time:
                    os.remove(log)



import warnings

import numpy as np
import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....rolling_mean import rolling_mean_2d
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    HEIGHT_VAR,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_cnom(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens CPR_NOM_1B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.JAXA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    reflectivity_var = "radarReflectivityFactor"
    doppler_velocity_var = "dopplerVelocity"

    # change the sign of the doppler velocity
    ds[doppler_velocity_var].values = ds[doppler_velocity_var].values * -1

    # Combine same dimensions
    ds = ds.squeeze()  # Removes dimensions of size 1
    ds = ds.rename({"phony_dim_14": "phony_dim_10"})
    ds = ds.rename({"phony_dim_15": "phony_dim_11"})
    # ds = ds.rename({'phony_dim_8': 'phony_dim_5'})

    # Rename content
    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="phony_dim_10",
        vertical_dim="phony_dim_11",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="binHeight",
        time_var="profileTime",
        elevation_var="surfaceElevation",
    )

    # Create plotting ready radar reflectivity and doppler velocity variables
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        values_ref = 10 * np.log10(ds[reflectivity_var].values).copy()
    values_ref = rolling_mean_2d(values_ref, 3, axis=1)
    values_ref = rolling_mean_2d(values_ref, 10, axis=0)

    new_radar_reflectivity_var = f"plot_{reflectivity_var}"
    ds[new_radar_reflectivity_var] = ds[reflectivity_var].copy()
    ds[new_radar_reflectivity_var].values = values_ref
    mask = values_ref < -27
    ds[new_radar_reflectivity_var].values[values_ref < -27] = np.nan
    ds = rename_var_info(
        ds,
        new_radar_reflectivity_var,
        "Radar reflectivity",
        "Radar reflectivity",
        "dBZ",
    )
    ds[new_radar_reflectivity_var] = ds[new_radar_reflectivity_var].assign_attrs(
        {
            "earthcarekit": "Added by earthcarekit: Intended for use in curtain plots only!"
        }
    )

    values_dvel = ds[doppler_velocity_var].values
    values_dvel = rolling_mean_2d(values_dvel, 3, axis=1)
    values_dvel = rolling_mean_2d(values_dvel, 10, axis=0)

    new_doppler_velocity_var = f"plot_{doppler_velocity_var}"
    ds[new_doppler_velocity_var] = ds[doppler_velocity_var].copy()
    ds[new_doppler_velocity_var].values = values_dvel
    ds[new_doppler_velocity_var].values[mask] = np.nan
    ds = rename_var_info(
        ds,
        new_doppler_velocity_var,
        "Doppler velocity",
        "Doppler velocity",
        "m/s",
    )
    ds[new_doppler_velocity_var] = ds[new_doppler_velocity_var].assign_attrs(
        {
            "earthcarekit": "Added by earthcarekit: Intended for use in curtain plots only!"
        }
    )

    # Remove nans from heights
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        ds[HEIGHT_VAR].values[np.isnan(ds[HEIGHT_VAR].values)] = np.nanmin(
            ds[HEIGHT_VAR].values
        )

    return ds



import numpy as np


def circular_lon_bbox(all_lons: np.ndarray) -> tuple[float, float]:
    """Compute smallest bounding box for longitudes, correctly handling dateline wrap."""
    lon = np.asarray(all_lons)
    lon = lon[~np.isnan(lon)]
    if lon.size == 0:
        return np.nan, np.nan

    lon360 = np.mod(lon, 360)
    lon360_sorted = np.sort(lon360)
    gaps = np.diff(np.concatenate([lon360_sorted, [lon360_sorted[0] + 360]]))
    max_gap_idx = np.argmax(gaps)

    lon_min = lon360_sorted[(max_gap_idx + 1) % len(lon360_sorted)]
    lon_max = lon360_sorted[max_gap_idx]

    # Convert back to [-180, 180]
    lon_min = ((lon_min + 180) % 360) - 180
    lon_max = ((lon_max + 180) % 360) - 180

    return lon_min, lon_max


def compute_bbox(*tracks: np.ndarray) -> list[float]:
    """
    Compute the minimal bounding box that contains all provided lat-lon tracks.

    Each track is either:
    - A tuple/list of (lat_sequence, lon_sequence)
    - Or a 2D array of shape (N, 2) where [:, 0] = lat, [:, 1] = lon

    Args:
        *tracks: One or more lat-lon coordinate sequences or 2D arrays.

    Returns:
        dict: Bounding box with lat_min, lat_max, lon_min, lon_max
    """
    all_lats = []
    all_lons = []

    for track in tracks:
        if isinstance(track, (tuple, list)) and len(track) == 2:
            lat, lon = np.asarray(track[0]), np.asarray(track[1])
        else:
            arr = np.asarray(track)
            if arr.ndim != 2 or arr.shape[1] != 2:
                raise ValueError("Each track must be (lat, lon) or Nx2 array.")
            lat, lon = arr[:, 0], arr[:, 1]

        all_lats.append(lat)
        all_lons.append(lon)

    all_lats_flat = np.concatenate(all_lats)
    all_lons_flat = np.concatenate(all_lons)

    lat_min = float(np.nanmin(all_lats_flat))
    lat_max = float(np.nanmax(all_lats_flat))
    lon_min, lon_max = circular_lon_bbox(all_lons_flat)

    extent = [
        lon_min,
        lon_max,
        lat_min,
        lat_max,
    ]
    return extent


def pad_bbox(
    extent: list[float],
    pad_lat: float | None = None,  # 2,
    pad_lon: float | None = None,  # 2,
    pad_lat_ratio: float | str | None = "1%",
    pad_lon_ratio: float | str | None = "1%",
) -> list[float]:
    lon_min = extent[0]
    lon_max = extent[1]
    lat_min = extent[2]
    lat_max = extent[3]

    _pad_lon: float = 0
    _pad_lat: float = 0

    if pad_lon is not None and pad_lon_ratio is not None:
        raise TypeError(
            f"both 'pad_lon' and 'pad_lon_ratio' are given but only one is required."
        )
    if pad_lon is not None:
        _pad_lon = pad_lon
    elif pad_lon_ratio is not None:
        lon_diff: float = float(np.abs(lon_max - lon_min))
        if isinstance(pad_lon_ratio, str):
            try:
                pad_lon_ratio = int(pad_lon_ratio[0:-1]) / 100
                if pad_lon_ratio < 0 or pad_lon_ratio > 1:
                    raise ValueError()
                _pad_lon = lon_diff * pad_lon_ratio
            except Exception as e:
                raise ValueError(
                    f"""invalid pad ratio string "{pad_lon_ratio}", expecting string like e.g. this: "75%"."""
                )
        elif isinstance(pad_lon_ratio, float):
            if pad_lon_ratio < 0 or pad_lon_ratio > 1:
                raise ValueError(
                    f"""invalid pad ratio float "{pad_lon_ratio}", expecting float between 0 and 1"""
                )
            _pad_lon = lon_diff * pad_lon_ratio

    if pad_lat is not None and pad_lat_ratio is not None:
        raise TypeError(
            f"both 'pad_lat' and 'pad_lat_ratio' are given but only one is required."
        )
    if pad_lat is not None:
        _pad_lat = pad_lat
    elif pad_lat_ratio is not None:
        lat_diff: float = float(np.abs(lat_max - lat_min))
        if isinstance(pad_lat_ratio, str):
            try:
                pad_lat_ratio = int(pad_lat_ratio[0:-1]) / 100
                if pad_lat_ratio < 0 or pad_lat_ratio > 1:
                    raise ValueError()
                _pad_lat = lat_diff * pad_lat_ratio
            except Exception as e:
                raise ValueError(
                    f"""invalid pad ratio string "{pad_lat_ratio}", expecting string like e.g. this: "75%"."""
                )
        elif isinstance(pad_lat_ratio, float):
            if pad_lat_ratio < 0 or pad_lat_ratio > 1:
                raise ValueError(
                    f"""invalid pad ratio float "{pad_lat_ratio}", expecting float between 0 and 1"""
                )
            _pad_lat = lat_diff * pad_lat_ratio

    extent = [
        lon_min - _pad_lon,
        lon_max + _pad_lon,
        lat_min - _pad_lat,
        lat_max + _pad_lat,
    ]
    return extent



from dataclasses import dataclass, field
from logging import Logger
from typing import Final

from ._request import get_request_json
from ._types import Entrypoint

_COLLECTION_SORT_ORDER: Final[list[str]] = [
    "EarthCAREL0L1Products",
    "EarthCAREL2Products",
    "JAXAL2Products",
    "EarthCAREAuxiliary",
    "EarthCAREL1InstChecked",
    "EarthCAREL2InstChecked",
    "JAXAL2InstChecked",
    "EarthCAREL1Validated",
    "EarthCAREL2Validated",
    "JAXAL2Validated",
    "EarthCAREXMETL1DProducts10",
    "EarthCAREOrbitData",
]
_COLLECTION_ORDER_MAP: Final[dict[str, int]] = {
    name: i for i, name in enumerate(_COLLECTION_SORT_ORDER)
}


@dataclass(order=True)
class EOCollection:
    _sort_index: int = field(init=False, repr=False)
    name: str
    url_queryables: str | None
    url_items: str | None

    def __post_init__(self):
        self._sort_index = _COLLECTION_ORDER_MAP.get(
            self.name, len(_COLLECTION_SORT_ORDER)
        )

    @property
    def sort_index(self):
        return self._sort_index

    @property
    def is_maap(self):
        return "_MAAP" in self.name


def _get_collections_url(entrypoint: Entrypoint, title: str | None = None) -> str:
    """Returns the URL of the collections search request."""
    if not isinstance(title, str) or title == "":
        title = "earthcare"
    return f"{entrypoint.value}/collections/?title={title}&limit=100&startRecord=1"


def get_available_collections(
    entrypoint: Entrypoint,
    title: str | None = None,
    logger: Logger | None = None,
) -> list[EOCollection]:
    """Requests EO collections from the Server and returns available ones."""
    url = _get_collections_url(entrypoint=entrypoint, title=title)

    data = get_request_json(url=url, logger=logger)

    eo_collections = []
    for collection in data.get("collections", []):
        collection_id = collection.get("id")
        url_queryables: str | None = None
        url_items: str | None = None

        for link in collection.get("links", []):
            _rel = link.get("rel")
            if _rel == "http://www.opengis.net/def/rel/ogc/1.0/queryables":
                url_queryables = link.get("href", None)
            if _rel == "items":
                url_items = link.get("href", None)
                if isinstance(url_items, str):
                    url_items = url_items.split("items")[0] + "items"
            if isinstance(url_queryables, str) and isinstance(url_items, str):
                break

        eoc = EOCollection(
            name=collection_id,
            url_queryables=url_queryables,
            url_items=url_items,
        )
        eo_collections.append(eoc)

    return eo_collections



from dataclasses import dataclass
from typing import Literal

import cartopy.crs as ccrs  # type: ignore
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from numpy.polynomial.legendre import leggauss
from numpy.typing import NDArray

from ...np_array_utils import flatten_array


def _get_lon_bounds_per_lat(
    nlon_equator: int,
    lat_centers: NDArray,
    reduced: bool = False,
) -> list[NDArray]:
    """
    Create longitude boudns per latitude.

    Args:
        nlon_equator (int): Number of longitude points at the equator.
        lat_centers (ndarray): Array of latitude centers.
        reduced (bool, optional): Whether to reduce number of longitudes near poles. Defaults to False.

    Returns:
        list[NDArray]: List of longitude bounds per given latitude center.
    """
    lon_bounds_list: list[NDArray] = []
    if not reduced:
        lon_bounds = np.linspace(-180, 180, nlon_equator + 1)
        lon_bounds_list = [lon_bounds] * len(lat_centers)
        return lon_bounds_list

    lat_radians = np.radians(lat_centers)

    for phi in lat_radians:
        scale = np.cos(phi)
        nlon_i = max(4, int(round(nlon_equator * scale)))

        lon_i = np.linspace(-180, 180, nlon_i + 1)
        lon_bounds_list.append(lon_i)

    return lon_bounds_list


def _get_regular_latitudes(nlat: int) -> tuple[NDArray, NDArray]:
    """Generate regular latitudes (equal angular distance)."""
    lat_centers = np.linspace(-90 + 90 / nlat, 90 - 90 / nlat, nlat)
    lat_bounds = np.linspace(-90, 90, nlat + 1)
    return lat_centers, lat_bounds


def _get_sinusoidal_latitudes(nlat: int) -> tuple[NDArray, NDArray]:
    """Generate sinusoidal latitudes (equal area spacing)."""
    mu_edges = np.linspace(-1, 1, nlat + 1)
    mu_centers = 0.5 * (mu_edges[:-1] + mu_edges[1:])
    lat_centers = np.degrees(np.arcsin(mu_centers))
    lat_bounds = np.degrees(np.arcsin(mu_edges))
    return lat_centers, lat_bounds


def _get_gaussian_latitudes(nlat: int) -> tuple[NDArray, NDArray]:
    """Generate Gauss-Legendre latitudes (quadrature points)."""
    mu, _ = leggauss(nlat)
    lat_centers = np.degrees(np.arcsin(mu))
    lat_bounds = np.zeros(nlat + 1)
    lat_bounds[1:-1] = 0.5 * (lat_centers[:-1] + lat_centers[1:])
    lat_bounds[0] = -90.0
    lat_bounds[-1] = 90.0
    return lat_centers, lat_bounds


@dataclass(frozen=True)
class SphericalGrid:
    lat_centers: NDArray
    lon_centers: NDArray | list[NDArray]
    lat_bounds: NDArray
    lon_bounds: NDArray | list[NDArray]
    is_reduced: bool

    def _get_cell_areas(self, radius: float = 6371e3, verbose: bool = True):
        """Compute the area of each grid cell."""
        lat_bounds = self.lat_bounds
        lon_bounds_list: list[NDArray]
        if self.is_reduced and isinstance(self.lon_bounds, list):
            lon_bounds_list = self.lon_bounds
        elif not self.is_reduced and isinstance(self.lon_bounds, np.ndarray):
            lon_bounds_list = [self.lon_bounds] * len(lat_bounds)
        else:
            raise TypeError(
                f"SphericalGrid: lon_bounds has invalid type {type(self.lon_bounds)} for is_reduced={self.is_reduced}.",
                "Expected type list[NDArray] for is_reduced=True and NDArray for is_reduced=False.",
            )
        areas = []
        for i in range(len(lat_bounds) - 1):
            phi1 = np.radians(lat_bounds[i])
            phi2 = np.radians(lat_bounds[i + 1])
            dphi = abs(np.sin(phi1) - np.sin(phi2))
            lon_bounds = lon_bounds_list[i]
            for j in range(len(lon_bounds) - 1):
                dlam = np.radians(lon_bounds[j + 1] - lon_bounds[j])
                area = (radius**2) * dphi * dlam
                areas.append(area)

        _areas = np.array(areas)
        if verbose:
            print(f"Mean area [km^2] {np.mean(_areas) / 1e6:.1f}")
            print(
                f"Area variation [%]: {100 * (np.max(_areas) - np.min(_areas)) / np.mean(_areas):.4f}",
            )
        return _areas

    def _preview(
        self,
        figsize=(10, 6),
        projection: (
            ccrs.Projection
            | Literal[
                "orthographic",
                "platecarree",
                "sinusoidal",
                "robinson",
                "eckertiv",
                "equalearth",
            ]
        ) = ccrs.PlateCarree(),
        color="tab:blue",
        lw=3,
    ):
        if isinstance(projection, str):
            if projection == "orthographic":
                projection = ccrs.Orthographic(central_latitude=45)
            elif projection == "platecarree":
                projection = ccrs.PlateCarree()
            elif projection == "sinusoidal":
                projection = ccrs.Sinusoidal()
            elif projection == "robinson":
                projection = ccrs.Robinson()
            elif projection == "eckertiv":
                projection = ccrs.EckertIV()
            elif projection == "equalearth":
                projection = ccrs.EqualEarth()

        fig, ax = plt.subplots(subplot_kw={"projection": projection}, figsize=figsize)

        ax.coastlines(resolution="110m", lw=1)  # type: ignore
        ax.gridlines(linestyle="dashed", color="black", alpha=0.5)  # type: ignore
        ax.set_global()  # type: ignore

        lats = self.lat_bounds
        lons = self.lon_bounds

        for i, lt in enumerate(lats):
            _ln = np.unique(flatten_array(lons))
            _lt = np.atleast_1d(lt).repeat(len(_ln))
            ax.plot(
                _ln,
                _lt,
                transform=ccrs.PlateCarree(),
                zorder=10,
                color=color,
                linewidth=lw,
            )

        if isinstance(lons, list):
            for i, ln in enumerate(lons):
                for j in range(len(ln) - 1):
                    _ln = np.array(ln[j]).repeat(2)
                    _lt = np.atleast_1d(lats[i : i + 2])
                    ax.plot(
                        _ln,
                        _lt,
                        transform=ccrs.PlateCarree(),
                        zorder=10,
                        color=color,
                        linewidth=lw,
                    )
        else:
            for i, ln in enumerate(lons):
                _lt = lats
                _ln = np.atleast_1d(ln).repeat(len(lats))
                ax.plot(
                    _ln,
                    _lt,
                    transform=ccrs.PlateCarree(),
                    zorder=10,
                    color=color,
                    linewidth=lw,
                )
        return fig, ax


def create_spherical_grid(
    nlat: int,
    nlon: int | None = None,
    reduced: bool = False,
    lat_spacing: Literal["regular", "sinusoidal", "gaussian"] = "regular",
) -> SphericalGrid:
    """
    Generate a spherical gird with regular, sinusoidal or gaussian latitudes and uniform or reduced longitudes.

    Args:
        nlat (int): Number of latitude.
        nlon (int | None): Nuber of longitudes at the equator. If None, set to `nlat * 2`. Defaults to None.
        reduced (bool): If True, reduces longitudes near poles using `~cos(latitude)` scaling. Defaults to False.
        lat_spacing ("regular" or "sinusoidal" or "gaussian", optional): Method used to place latitudes. Defaults to "regular".

    Returns:
        SphericalGrid: A container storing

        - lat_centers: A `numpy.array` of latitude bin centers.
        - lon_centers: A `numpy.array` of longitude bin centers or if `is_reduced=True` a list of `numpy.array` per latitude center.
        - lat_bounds: A `numpy.array` of latitude bin bounds.
        - lon_bounds: A `numpy.array` of longitude bin bounds or if `is_reduced=True` a list of `numpy.array` per latitude bound.
        - is_reduced (bool): Whether the grid has reduced number of longitudes near the poles.
    """

    if nlon is None:
        nlon = nlat + nlat

    if lat_spacing == "regular":
        lat_c, lat_b = _get_regular_latitudes(nlat)
    elif lat_spacing == "sinusoidal":
        lat_c, lat_b = _get_sinusoidal_latitudes(nlat)
    elif lat_spacing == "gaussian":
        lat_c, lat_b = _get_gaussian_latitudes(nlat)
    else:
        raise ValueError("grid_type must be 'regular', 'gaussian', or 'sinusoidal'")

    lon_b_list = _get_lon_bounds_per_lat(nlon, lat_c, reduced)

    lon_b: NDArray | list[NDArray]
    if reduced:
        lon_b = lon_b_list
        lon_c = [0.5 * (lb[:-1] + lb[1:]) for lb in lon_b_list]
    else:
        lon_b = lon_b_list[0]
        lon_c = 0.5 * (lon_b[:-1] + lon_b[1:])

    return SphericalGrid(
        lat_centers=lat_c,
        lon_centers=lon_c,
        lat_bounds=lat_b,
        lon_bounds=lon_b,
        is_reduced=reduced,
    )



import warnings
from dataclasses import asdict, dataclass
from typing import Iterable, Literal, Tuple, TypeAlias

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import ArrayLike, NDArray

from ..constants import ACROSS_TRACK_DISTANCE, FROM_TRACK_DISTANCE
from ..rolling_mean import rolling_mean_2d
from ..time import TimeRangeLike, TimestampLike, to_timestamps, validate_time_range
from ..typing import DistanceRangeLike
from .across_track_distance import add_across_track_distance


@dataclass
class SwathData:
    values: NDArray
    time: NDArray
    latitude: NDArray
    longitude: NDArray
    nadir_index: int
    color: str | None = None
    label: str | None = None
    units: str | None = None
    platform: str | None = None

    def __post_init__(self):

        if isinstance(self.values, Iterable):
            self.values = np.atleast_2d(self.values)
        if isinstance(self.time, Iterable):
            self.time = pd.to_datetime(np.asarray(self.time)).to_numpy()
        if isinstance(self.latitude, Iterable):
            self.latitude = np.atleast_2d(self.latitude)
        if isinstance(self.longitude, Iterable):
            self.longitude = np.atleast_2d(self.longitude)

        if len(self.values.shape) == 2:
            if self.values.shape[0] != self.latitude.shape[0]:
                raise ValueError(
                    f"along_track dimension must match for `values` {self.values.shape} and `latitude` {self.latitude.shape}"
                )
            if self.values.shape[0] != self.longitude.shape[0]:
                raise ValueError(
                    f"along_track dimension must match for `values` {self.values.shape} and `longitude` {self.longitude.shape}"
                )
            value_dims = ("along_track", "cross_track")
        elif len(self.values.shape) == 3:
            if self.values.shape[1] != self.latitude.shape[0]:
                raise ValueError(
                    f"along_track dimension must match for `values` {self.values.shape} and `latitude` {self.latitude.shape}"
                )
            if self.values.shape[1] != self.longitude.shape[0]:
                raise ValueError(
                    f"along_track dimension must match for `values` {self.values.shape} and `longitude` {self.longitude.shape}"
                )
            value_dims = ("cross_track", "along_track", "rbg")
        else:
            raise ValueError(
                f"`values` but have 2 or 3 dims, but got {self.values.shape}"
            )

        if self.latitude.shape != self.longitude.shape:
            raise ValueError(
                f"`latitude` {self.latitude.shape} and `longitude` {self.longitude.shape} must have same shape"
            )

        ds = xr.Dataset(
            data_vars={
                "values": (value_dims, self.values),
                "time": (("along_track"), self.time),
                "latitude": (("along_track", "cross_track"), self.latitude),
                "longitude": (("along_track", "cross_track"), self.longitude),
            }
        )
        ds = add_across_track_distance(ds, self.nadir_index, "latitude", "longitude")
        self.across_track_distance = ds[ACROSS_TRACK_DISTANCE].values
        self.from_track_distance = ds[FROM_TRACK_DISTANCE].values

    @property
    def shape(self):
        return self.values.shape

    @property
    def track_latitude(self):
        return self.latitude[:, self.nadir_index]

    @property
    def track_longitude(self):
        return self.longitude[:, self.nadir_index]

    def select_time_range(
        self,
        time_range: TimeRangeLike | None,
    ) -> "SwathData":
        """Retruns only data within the specified `time_range`."""
        if not isinstance(self.time, np.ndarray):
            raise ValueError(
                f"{SwathData.__name__}.{self.select_time_range.__name__}() missing `time` data"
            )
        if time_range is None:
            return self

        time_range = validate_time_range(time_range)

        times = to_timestamps(self.time)
        mask = np.logical_and(time_range[0] <= times, times <= time_range[1])

        if (
            len(self.values.shape) >= 2
            and self.values.shape[0] != mask.shape[0]
            and self.values.shape[1] == mask.shape[0]
        ):
            sel_values = self.values[:, mask]
        else:
            sel_values = self.values[mask]
        sel_time = self.time[mask]
        sel_latitude = self.latitude[mask]
        sel_longitude = self.longitude[mask]

        return SwathData(
            values=sel_values,
            time=sel_time,
            latitude=sel_latitude,
            longitude=sel_longitude,
            nadir_index=self.nadir_index,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
        )

    def select_from_track_range(
        self,
        from_track_range: DistanceRangeLike | None,
    ) -> "SwathData":
        """Retruns only data within the specified `from_track_range`."""
        if from_track_range is None:
            return self

        mask = np.logical_and(
            from_track_range[0] <= self.from_track_distance,
            self.from_track_distance <= from_track_range[1],
        )

        sel_values = self.values[:, mask]
        sel_time = self.time
        sel_latitude = self.latitude[:, mask]
        sel_longitude = self.longitude[:, mask]
        new_nadir_index = np.sum(mask[: self.nadir_index])

        return SwathData(
            values=sel_values,
            time=sel_time,
            latitude=sel_latitude,
            longitude=sel_longitude,
            nadir_index=new_nadir_index,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
        )



from typing import Any, Literal, Sequence, Type, TypeAlias

import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from numpy.typing import NDArray

from ..color.color import Color, ColorLike

ClassIDInt: TypeAlias = int
ClassLabelStr: TypeAlias = str


def plot_1d_integer_flag(
    ax: Axes,
    data: Sequence | NDArray,
    classes: Sequence[ClassIDInt] | dict[ClassIDInt, ClassLabelStr],
    x: Sequence | NDArray | None = None,
    ax2: Axes | None = None,
    color: ColorLike | Sequence[ColorLike | None] | None = None,
    alpha: float | Sequence[float] = 1.0,
    linewidth: float = 10.0,
    show_hline: bool | list[bool] = True,
    hline_color: ColorLike | Sequence[ColorLike | None] | None = "black",
    hline_alpha: float | Sequence[float] = 1.0,
    hline_linestyle: str | Sequence[str] = "solid",
    hline_linewidth: float | Sequence[float] = 1.5,
    tick_color: ColorLike | Sequence[ColorLike | None] | None = "black",
    label_color: ColorLike | Sequence[ColorLike | None] | None = "black",
    yaxis_position: str | Literal["top", "bottom", "both", "default", "none"] = "left",
    ax_label: str | None = None,
    zorder: int | float = 2,
) -> None:

    class_labels_dict: dict[ClassIDInt, ClassLabelStr] | None = None
    if not isinstance(classes, dict):
        value_labels_dict = {int(c): str(int(c)) for c in classes}
    else:
        value_labels_dict = {int(k): str(int(k)) for k in classes.keys()}
        class_labels_dict = {int(k): str(v) for k, v in classes.items()}

    def _validate_list(
        var: Any,
        var_name: str,
        non_list_types: tuple[Type] | None = (str,),
        list_types: tuple = (Sequence, np.ndarray),
    ) -> bool:
        is_list = isinstance(var, list_types)
        is_list_exception = (
            False if non_list_types is None else isinstance(var, non_list_types)
        )

        is_list = is_list and not is_list_exception

        if is_list:
            if len(var) != len(classes):
                raise ValueError(
                    f"{var_name} list must have the same length as classes: len({var_name})={len(var)} != {len(classes)=}"
                )

        return is_list

    if _validate_list(color, "color"):
        color = [Color.from_optional(c) for c in color]  # type: ignore
    else:
        color = Color.from_optional(color)  # type: ignore
        color = [color] * len(value_labels_dict)

    if _validate_list(alpha, "alpha"):
        alpha = [a for a in alpha]  # type: ignore
    else:
        alpha = [alpha] * len(value_labels_dict)  # type: ignore

    if _validate_list(show_hline, "show_hline"):
        show_hline = [shl for shl in show_hline]  # type: ignore
    else:
        show_hline = [show_hline] * len(value_labels_dict)  # type: ignore

    if _validate_list(hline_color, "color_hline"):
        hline_color = [Color.from_optional(c) for c in hline_color]  # type: ignore
    else:
        hline_color = Color.from_optional(hline_color)  # type: ignore
        hline_color = [hline_color] * len(value_labels_dict)

    if _validate_list(hline_alpha, "hline_alpha"):
        hline_alpha = [a for a in hline_alpha]  # type: ignore
    else:
        hline_alpha = [hline_alpha] * len(value_labels_dict)  # type: ignore

    if _validate_list(hline_linestyle, "hline_linestyle"):
        hline_linestyle = [ls for ls in hline_linestyle]
    else:
        hline_linestyle = [hline_linestyle] * len(value_labels_dict)  # type: ignore

    if _validate_list(hline_linewidth, "hline_linewidth"):
        hline_linewidth = [lw for lw in hline_linewidth]  # type: ignore
    else:
        hline_linewidth = [hline_linewidth] * len(value_labels_dict)  # type: ignore

    if _validate_list(tick_color, "ticks_color"):
        tick_color = [Color.from_optional(c) for c in tick_color]  # type: ignore
    else:
        tick_color = Color.from_optional(tick_color)  # type: ignore
        tick_color = [tick_color] * len(value_labels_dict)

    if _validate_list(label_color, "label_color"):
        label_color = [Color.from_optional(c) for c in label_color]  # type: ignore
    else:
        label_color = Color.from_optional(label_color)  # type: ignore
        label_color = [label_color] * len(value_labels_dict)

    color_dict: dict[int, Color | None] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), color)
    }
    alpha_dict: dict[int, float] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), alpha)
    }
    show_hline_dict: dict[int, bool] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), show_hline)
    }
    hline_color_dict: dict[int, Color | None] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), hline_color)
    }
    hline_alpha_dict: dict[int, float] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), hline_alpha)
    }
    hline_linestyle_dict: dict[int, str] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), hline_linestyle)
    }
    hline_linewidth_dict: dict[int, float] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), hline_linewidth)
    }
    tick_color_dict: dict[int, Color | None] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), tick_color)
    }
    label_color_dict: dict[int, Color | None] = {
        k: c for k, c in zip(list(value_labels_dict.keys()), label_color)
    }

    class_series = np.asarray(data)
    class_series = np.round(class_series)
    unique_classes = list(value_labels_dict.keys())
    unique_classes = sorted(unique_classes)
    class_categorical = pd.Categorical(
        class_series, categories=unique_classes, ordered=True
    )
    value_labels: list[str] = [value_labels_dict[c] for c in unique_classes]
    class_labels: list[str] | None = (
        []
        if class_labels_dict is None
        else [class_labels_dict[c] for c in unique_classes]
    )

    if x is None:
        x = np.arange(class_series.shape[0])
    else:
        x = np.asarray(x)

    _final_colors = {}
    _final_hline_colors = {}

    for i, class_id in enumerate(unique_classes):
        class_name = (
            value_labels_dict[class_id]
            if class_id in value_labels_dict
            else str(class_id)
        )
        _color = color_dict[class_id]
        _alpha = alpha_dict[class_id]
        _show_hline = show_hline_dict[class_id]
        _hline_color = hline_color_dict[class_id]
        _hline_alpha = hline_alpha_dict[class_id]
        _hline_linestyle = hline_linestyle_dict[class_id]
        _hline_linewidth = hline_linewidth_dict[class_id]
        mask = class_categorical != class_id
        y = np.array(class_categorical.codes).astype(float).copy()
        y[mask] = np.nan
        if _show_hline:
            hl = ax.axhline(
                y=i,
                color=_hline_color,
                alpha=_hline_alpha,
                zorder=1,
                linestyle=_hline_linestyle,
                linewidth=_hline_linewidth,
            )
            _final_hline_colors[class_id] = hl.get_color()
        (p,) = ax.plot(
            x,
            y,
            linewidth=linewidth,
            markersize=linewidth * 0.9,
            marker="",
            color=_color,
            alpha=_alpha,
            solid_capstyle="butt",
            zorder=zorder,
            label=class_name,
        )
        _final_colors[class_id] = p.get_color()

    ax.set_yticks(np.arange(len(unique_classes)), value_labels)
    vpad = 0.5
    ax.set_ylim(-vpad, len(unique_classes) + (vpad - 1.0))

    if isinstance(class_labels, list) and len(class_labels) > 0:
        if not isinstance(ax2, Axes):
            ax2 = ax.twinx()
        ax2.set_ylim(ax.get_ylim())

        ax.yaxis.set_ticks_position(yaxis_position)  # type: ignore
        yaxis_position2 = "right" if yaxis_position == "left" else "left"

        ax2.yaxis.set_ticks_position(yaxis_position2)  # type: ignore
        ax2.set_yticks(np.arange(len(unique_classes)), class_labels)
        for i, class_id in enumerate(unique_classes):
            _color = label_color_dict[class_id]
            _side_idx_offset = 0 if yaxis_position2 == "left" else 1
            ax2.get_yticklines()[i * 2 + _side_idx_offset].set_markeredgecolor(
                "#ffffff00"
            )
            _ticks_color = tick_color_dict[class_id]
            ax2.get_yticklabels()[i].set_color(_color)  # type: ignore

    for i, class_id in enumerate(unique_classes):
        if show_hline_dict[class_id]:
            line_color = _final_hline_colors[class_id]
            _hline_linewidth = hline_linewidth_dict[class_id]
            _side_idx_offset = 0 if yaxis_position == "left" else 1
            ax.get_yticklines()[i * 2 + _side_idx_offset].set_markeredgecolor(
                line_color
            )
            ax.get_yticklines()[i * 2 + _side_idx_offset].set_markeredgewidth(
                _hline_linewidth
            )

        _ticks_color = tick_color_dict[class_id]
        ax.get_yticklabels()[i].set_color(_ticks_color)  # type: ignore

    if isinstance(ax_label, str):
        ax.set_ylabel(ax_label)



import os
from logging import Logger


def _parse_dirpath(
    dirpath: str | None,
    logger: Logger | None = None,
) -> str | None:
    if not isinstance(dirpath, str):
        return None

    dirpath = os.path.abspath(dirpath)

    if not os.path.exists(dirpath):
        raise FileNotFoundError(f"No directory found at <{dirpath}>.")

    return dirpath


def parse_path_to_data(
    path_to_data: str | None,
    logger: Logger | None = None,
) -> str | None:
    try:
        return _parse_dirpath(dirpath=path_to_data, logger=logger)
    except FileNotFoundError as e:
        raise FileNotFoundError(
            f"No directory found at path_to_data: <{path_to_data}>."
        )


def parse_path_to_imgs(
    path_to_imgs: str | None,
    logger: Logger | None = None,
) -> str | None:
    try:
        return _parse_dirpath(dirpath=path_to_imgs, logger=logger)
    except FileNotFoundError as e:
        raise FileNotFoundError(
            f"No directory found at path_to_imgs: <{path_to_imgs}>."
        )



from .config import (
    get_config,
    read_config,
    set_config,
    set_config_maap_token,
    set_config_to_maap,
    set_config_to_oads,
)
from .ground_sites import GroundSite, get_ground_site
from .np_array_utils import ismonotonic, isndarray
from .profile_data.profile_data import ProfileData
from .read import *
from .rolling_mean import *
from .set import all_in
from .swath_data.swath_data import SwathData
from .xarray_utils import filter_index, filter_latitude, filter_radius, filter_time



import os
from typing import Any, Callable, Sequence, Type

import numpy as np
import pandas as pd

from ...config import ECKConfig, read_config
from ...time import TimestampLike, to_timestamp
from ..search import search_files_by_regex
from ._header_file import read_hdr_fixed_header
from .file_info import (
    FileAgency,
    FileInfoEnum,
    FileLatency,
    FileType,
    ProductDataFrame,
    format_frame_id,
    format_orbit_and_frame,
    format_orbit_number,
    get_product_info,
    get_product_infos,
    validate_baseline,
)


def _to_file_info_list(
    input: str | Sequence[str] | None, file_info_enum: Type[FileInfoEnum]
) -> list[FileInfoEnum]:
    """Converts input string(s) to list of `FileInfoEnum` instances."""
    if isinstance(input, str):
        return [file_info_enum.from_input(x).value for x in np.atleast_1d(input)]
    elif isinstance(input, Sequence):
        return [file_info_enum.from_input(x).value for x in input]
    return []


def _format_input(
    input: Any | None,
    file_types: list | None = None,
    default_input: str | None = None,
    format_func: Callable | None = None,
) -> list:
    """
    Applies format function to input(s) and depeding on if a list of file types
    is given checks if the number of inputs matches the number of outputs.
    """
    if input is not None:
        input = list(np.atleast_1d(input))

        if callable(format_func):
            input = [format_func(x) for x in input]

        if isinstance(file_types, list):
            if len(input) == 1:
                input = [input[0]] * len(file_types)
            elif len(input) == 0 and isinstance(default_input, str):
                input = [default_input] * len(file_types)
            elif len(input) != len(file_types):
                raise ValueError(
                    f"Number of inputs ({len(input)}: {input}) given is greater than 1 and does not match number of given file types ({len(file_types)}: {file_types})"
                )
    else:
        if isinstance(default_input, str) and isinstance(file_types, list):
            input = [default_input] * len(file_types)
        else:
            input = []
    return input


def _list_to_regex(
    input: list[str] | list[FileInfoEnum],
    default_input: str,
) -> str:
    """Joins strings in list to a regular expression string."""
    if len(input) == 0:
        return default_input
    return f"({'|'.join(input)})"


def _check_product_contains_timestamp(
    filepath: str,
    timestamp: TimestampLike,
) -> bool:
    """Checks if product may contain timestamp."""
    product_info = get_product_info(filepath)
    timestamp = to_timestamp(timestamp)
    if (
        product_info.start_sensing_time <= timestamp
        and product_info.start_sensing_time + pd.Timedelta(minutes=13) >= timestamp
    ):

        hdr_filepath = os.path.join(
            os.path.dirname(filepath), os.path.basename(filepath).split(".")[0] + ".HDR"
        )
        if os.path.exists(hdr_filepath):
            hdr = read_hdr_fixed_header(hdr_filepath)
            timestamp = to_timestamp(timestamp)
            if isinstance(hdr.validity_start, pd.Timestamp) and isinstance(
                hdr.validity_stop, pd.Timestamp
            ):
                if hdr.validity_start <= timestamp and hdr.validity_stop >= timestamp:
                    return True
        else:
            return True
    return False


def filter_time_range(
    df: ProductDataFrame,
    start_time: TimestampLike | None = None,
    end_time: TimestampLike | None = None,
    time_column: str = "start_sensing_time",
) -> ProductDataFrame:
    df.validate_columns()

    if start_time is None and end_time is None:
        return df

    df = df.sort_values(time_column).reset_index(drop=True)

    start_mask = pd.Series(True, index=df.index)
    if start_time is not None:
        start_time = to_timestamp(start_time)
        start_mask = df[time_column] >= start_time

    end_mask = pd.Series(True, index=df.index)
    if end_time is not None:
        end_time = to_timestamp(end_time)
        end_mask = df[time_column] <= end_time

    mask = np.logical_and(start_mask, end_mask)

    df_filtered = df[mask].copy()
    df_filtered.validate_columns()

    if df_filtered.empty:
        return df_filtered

    return df_filtered


def search_pattern(
    file_type: str | Sequence[str] | None = None,
    agency: str | Sequence[str] | None = None,
    latency: str | Sequence[str] | None = None,
    timestamp: TimestampLike | Sequence[TimestampLike] | None = None,
    baseline: str | Sequence[str] | None = None,
    orbit_and_frame: str | Sequence[str] | None = None,
    orbit_number: int | str | Sequence[int | str] | None = None,
    frame_id: str | Sequence[str] | None = None,
):
    """
    Searches for EarthCARE product files matching given metadata filters.

    Args:
        config_filepath (str , optional): Path to a `config.toml` file. Defaults to the default configuration file.
        file_type (str or Sequence[str], optional): Product file type(s) to match.
        agency (str or Sequence[str], optional): Producing agency or agencies (e.g. "ESA" or "JAXA").
        latency (str or Sequence[str], optional): Data latency level(s).
        timestamp (TimestampLike or Sequence, optional): Timestamp(s) included in the product's time coverage.
        baseline (str or Sequence[str], optional): Baseline version(s).
        orbit_and_frame (str or Sequence[str], optional): Orbit and frame identifiers.
        orbit_number (int, str, or Sequence, optional): Orbit number(s).
        frame_id (str or Sequence[str], optional): Frame identifier(s).
        filename (str or Sequence[str], optional): Specific filename(s) or regular expression patterns to match.

    Returns:
        ProductDataFrame: Filtered list of matching product files as a `xarray.DataFrame`.

    Raises:
        FileNotFoundError: If root directory does not exist.
    """
    mission_id = "ECA"

    file_type = _to_file_info_list(file_type, FileType)
    baseline = _format_input(
        baseline,
        file_types=file_type,
        default_input="..",
        format_func=validate_baseline,
    )
    baseline_and_file_type_list = [f"{bl}_{ft}" for bl, ft in zip(baseline, file_type)]
    baseline_and_file_type = _list_to_regex(
        baseline_and_file_type_list, ".._..._..._.."
    )

    agency = _to_file_info_list(agency, FileAgency)
    agency = _list_to_regex(agency, ".")

    latency = _to_file_info_list(latency, FileLatency)
    latency = _list_to_regex(latency, ".")

    timestamp = _format_input(timestamp, format_func=to_timestamp)

    orbit_and_frame = _format_input(orbit_and_frame, format_func=format_orbit_and_frame)
    orbit_and_frame = _list_to_regex(orbit_and_frame, "." * 6)

    orbit_number = _format_input(orbit_number, format_func=format_orbit_number)
    orbit_number = _list_to_regex(orbit_number, "." * 5)

    frame_id = _format_input(frame_id, format_func=format_frame_id)
    frame_id = _list_to_regex(frame_id, ".")

    oaf_list = []
    oaf = ""
    if orbit_number != "." * 5:
        oaf_list.append(orbit_number)
    if frame_id != ".":
        oaf_list.append(frame_id)
    if orbit_number != "." * 5 or frame_id != ".":
        oaf = f"{orbit_number}{frame_id}"

    if oaf == "":
        oaf = orbit_and_frame
    elif oaf != "" and orbit_and_frame != "." * 6:
        oaf = f"(({oaf})|{orbit_and_frame})"

    pattern = f".*{mission_id}_{agency}{latency}{baseline_and_file_type}_........T......Z_........T......Z_{oaf}.h5"

    return pattern


def search_product(
    root_dirpath: str | None = None,
    config: str | ECKConfig | None = None,
    file_type: str | Sequence[str] | None = None,
    agency: str | Sequence[str] | None = None,
    latency: str | Sequence[str] | None = None,
    timestamp: TimestampLike | Sequence[TimestampLike] | None = None,
    baseline: str | Sequence[str] | None = None,
    orbit_and_frame: str | Sequence[str] | None = None,
    orbit_number: int | str | Sequence[int | str] | None = None,
    frame_id: str | Sequence[str] | None = None,
    filename: str | Sequence[str] | None = None,
    start_time: TimestampLike | None = None,
    end_time: TimestampLike | None = None,
) -> ProductDataFrame:
    """
    Searches for EarthCARE product files matching given metadata filters.

    Args:
        root_dirpath (str, optional): Root directory to search. Defaults to directory given in a configuration file.
        config (str | ECKConfig | None , optional): Path to a `config.toml` file or a ECKConfig instance. Defaults to the default configuration file path.
        file_type (str | Sequence[str], optional): Product file type(s) to match.
        agency (str | Sequence[str], optional): Producing agency or agencies (e.g. "ESA" or "JAXA").
        latency (str | Sequence[str], optional): Data latency level(s).
        timestamp (TimestampLike | Sequence, optional): Timestamp(s) included in the product's time coverage.
        baseline (str | Sequence[str], optional): Baseline version(s).
        orbit_and_frame (str | Sequence[str], optional): Orbit and frame identifiers.
        orbit_number (int, str, | Sequence, optional): Orbit number(s).
        frame_id (str | Sequence[str], optional): Frame identifier(s).
        filename (str | Sequence[str], optional): Specific filename(s) or regular expression patterns to match.
        start_time (TimestampLike, optional): First timestamp included in the product's time coverage.
        end_time (TimestampLike, optional): Last timestamp included in the product's time coverage.

    Returns:
        ProductDataFrame: Filtered list of matching product files as a `pandas.DataFrame`-based object.

    Raises:
        FileNotFoundError: If root directory does not exist.
    """
    if not isinstance(root_dirpath, str):
        if isinstance(config, ECKConfig):
            root_dirpath = config.path_to_data
        else:
            root_dirpath = read_config(config).path_to_data

    if not os.path.exists(root_dirpath):
        raise FileNotFoundError(f"Given root directory does not exist: {root_dirpath}")

    mission_id = "ECA"

    if isinstance(file_type, str):
        file_type = [file_type]
    if isinstance(file_type, Sequence):
        _baseline: list[str] = []
        _file_type: list[str] = []
        for i, ft in enumerate(file_type):
            if isinstance(ft, str):
                _parts = ft.split(":")
                if len(_parts) == 2:
                    _file_type.append(_parts[0])
                    _baseline.append(_parts[1])
                    continue
            _file_type.append(ft)
            if isinstance(baseline, str):
                _baseline.append(baseline)
            elif isinstance(baseline, Sequence):
                try:
                    _baseline.append(baseline[i])
                except IndexError as e:
                    raise IndexError(e, f"given baseline list is too small")
            else:
                _baseline.append("latest")
        file_type = _file_type
        baseline = _baseline
    file_type = _to_file_info_list(file_type, FileType)
    baseline = _format_input(
        baseline,
        file_types=file_type,
        default_input="..",
        format_func=validate_baseline,
    )
    baseline_and_file_type_list = [f"{bl}_{ft}" for bl, ft in zip(baseline, file_type)]
    baseline_and_file_type = _list_to_regex(
        baseline_and_file_type_list, ".._..._..._.."
    )

    agency = _to_file_info_list(agency, FileAgency)
    agency = _list_to_regex(agency, ".")

    latency = _to_file_info_list(latency, FileLatency)
    latency = _list_to_regex(latency, ".")

    timestamp = _format_input(timestamp, format_func=to_timestamp)
    _start_time = [] if start_time is None else [to_timestamp(start_time)]
    _end_time = [] if end_time is None else [to_timestamp(end_time)]
    timestamp = timestamp + _start_time + _end_time

    orbit_and_frame = _format_input(orbit_and_frame, format_func=format_orbit_and_frame)
    orbit_and_frame = _list_to_regex(orbit_and_frame, "." * 6)

    orbit_number = _format_input(orbit_number, format_func=format_orbit_number)
    orbit_number = _list_to_regex(orbit_number, "." * 5)

    frame_id = _format_input(frame_id, format_func=format_frame_id)
    frame_id = _list_to_regex(frame_id, ".")

    oaf_list = []
    oaf = ""
    if orbit_number != "." * 5:
        oaf_list.append(orbit_number)
    if frame_id != ".":
        oaf_list.append(frame_id)
    if orbit_number != "." * 5 or frame_id != ".":
        oaf = f"{orbit_number}{frame_id}"

    if oaf == "":
        oaf = orbit_and_frame
    elif oaf != "" and orbit_and_frame != "." * 6:
        oaf = f"(({oaf})|{orbit_and_frame})"

    pattern = f".*{mission_id}_{agency}{latency}{baseline_and_file_type}_........T......Z_........T......Z_{oaf}.h5"

    # pattern = search_pattern(
    #     file_type=file_type,
    #     agency=agency,
    #     latency=latency,
    #     timestamp=timestamp,
    #     baseline=baseline,
    #     orbit_and_frame=orbit_and_frame,
    #     orbit_number=orbit_number,
    #     frame_id=frame_id,
    # )

    if pattern == ".*ECA_...._..._..._.._........T......Z_........T......Z_.......h5":
        files = []
    else:
        files = search_files_by_regex(root_dirpath, pattern)

    if isinstance(filename, str) or isinstance(filename, Sequence):
        if isinstance(filename, str):
            filename = [filename]
        _get_pattern = lambda fn: f".*{os.path.basename(fn).replace('.h5', '')}.*.h5"
        filename = [_get_pattern(fn) for fn in filename]
    elif filename is None:
        filename = []
    else:
        raise TypeError(
            f"Given filename has invalid type ({type(filename)}: {filename})"
        )

    for fn in filename:
        new_files = search_files_by_regex(root_dirpath, fn)
        files.extend(new_files)

    # Remove duplicates
    files = list(set(files))

    old_files = files.copy()
    if len(timestamp) > 0:
        files = []
        for t in timestamp:
            new_files = [
                f for f in old_files if _check_product_contains_timestamp(f, t)
            ]
            if len(new_files) > 0:
                files.extend(new_files)

    pdf = get_product_infos(files)

    if start_time is not None or end_time is not None:
        _pdf = get_product_infos(old_files)
        _pdf = filter_time_range(_pdf, start_time=start_time, end_time=end_time)

        if not pdf.empty and not _pdf.empty:
            pdf = ProductDataFrame(pd.concat([pdf, _pdf], ignore_index=True))
        elif not _pdf.empty:
            pdf = _pdf

    pdf = pdf.sort_values(by=["orbit_and_frame", "file_type", "start_processing_time"])
    pdf = pdf.drop_duplicates()
    pdf = pdf.reset_index(drop=True)

    pdf.validate_columns()
    return pdf



from typing import Literal


def get_relative_luminance(rgb: tuple[int, int, int]) -> float:
    """
    Compute WCAG 2.0 relative luminance for an RGB color.

    See https://www.w3.org/TR/WCAG20/#relativeluminancedef
    """
    r, g, b = rgb

    def srgb_from_8bit(c):
        c = c / 255.0
        if c <= 0.03928:
            return c / 12.92
        else:
            return ((c + 0.055) / 1.055) ** 2.4

    R = srgb_from_8bit(r)
    G = srgb_from_8bit(g)
    B = srgb_from_8bit(b)

    L = 0.2126 * R + 0.7152 * G + 0.0722 * B

    return L


def get_contrast_ratio(l1: float, l2: float) -> float:
    """
    WCAG 2.0 contrast ratio between two luminances.

    See https://www.w3.org/TR/WCAG20/#contrast-ratiodef
    """
    L1, L2 = sorted([l1, l2])
    return (L2 + 0.05) / (L1 + 0.05)


def get_best_bw_contrast_color(rgb: tuple[int, int, int]) -> Literal["black", "white"]:
    """
    Return 'black' or 'white' according to WCAG 2.0 contrast ratio.

    See https://www.w3.org/TR/WCAG20/
    """
    L1 = get_relative_luminance(rgb)
    L_white = 1.0
    L_black = 0.0
    contrast_white = get_contrast_ratio(L1, L_white)
    contrast_black = get_contrast_ratio(L1, L_black)

    if contrast_white > contrast_black:
        return "white"
    else:
        return "black"



from typing import Final

import numpy as np
import xarray as xr
from scipy.spatial import cKDTree  # type: ignore

from ...constants import (
    ACROSS_TRACK_DIM,
    ALONG_TRACK_DIM,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
    TIME_VAR,
)
from ...geo import sequence_geo_to_ecef
from ._generic import read_product

_SKIP_VARS: Final[list[str]] = [
    "filename",
    "file_type",
    "frame_id",
    "orbit_number",
    "orbit_and_frame",
    "baseline",
    "sensing_start_time",
    "processing_start_time",
]


def rebin_msi_to_jsg(
    ds_msi: xr.Dataset | str,
    ds_xjsg: xr.Dataset | str,
    vars: list[str] | None = None,
    k: int = 4,
    eps: float = 1e-12,
    lat_var: str = SWATH_LAT_VAR,
    lon_var: str = SWATH_LON_VAR,
    time_var: str = TIME_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    across_track_dim: str = ACROSS_TRACK_DIM,
    lat_var_xjsg: str = SWATH_LAT_VAR,
    lon_var_xjsg: str = SWATH_LON_VAR,
    time_var_xjsg: str = TIME_VAR,
    along_track_dim_xjsg: str = ALONG_TRACK_DIM,
    across_track_dim_xjsg: str = ACROSS_TRACK_DIM,
) -> xr.Dataset:
    """
    Rebins variables from an MSI product dataset onto the geo-spacial lat/lon grid given by the related AUX_JSG_1D dataset.

    This function interpolates selected variables from `ds_msi` onto the JSG grid from `ds_xjsg`
    using quick kd-tree nearest-neighbor search with `scipy.spatial.cKDTree` followed
    by averaging the `k`-nearest points using inverse distance weighting. The resulting dataframe
    matches the along- and across-track resolution of `ds_xjsg`.

    Args:
        ds_msi (xr.Dataset | str): The source MSI dataset (e.g., MSI_RGR_1C, MSI_COP_2A, ...).
        ds_xjsg (xr.Dataset | str): The target XJSG dataset.
        vars (list[str] | None, optional): List of variable names from `ds_msi` to rebin.
            If None, all data variables are considered. Defaults to None.
        k (int, optional): Number of nearest geo-spacial neighbors to include in the kd-tree search.
            Defaults to 4.
        eps (float, optional): Numerical threshold to avoid division by zero in distance calculations during the kd-tree search.
            Defaults to 1e-12.

    Returns:
        xr.Dataset: The MSI dataset with variables rebinned to the JSG grid.
    """

    def _read_msi() -> xr.Dataset:
        if isinstance(ds_msi, str):
            return read_product(ds_msi)
        return ds_msi

    def _read_xjsg() -> xr.Dataset:
        if isinstance(ds_xjsg, str):
            return read_product(ds_xjsg)
        return ds_xjsg

    with (
        _read_msi() as ds_msi,
        _read_xjsg() as ds_xjsg,
    ):
        if vars is None:
            vars = [str(v) for v in ds_msi.variables]
        else:
            for var in vars:
                if var not in ds_msi.variables:
                    present_vars = [str(v) for v in ds_msi.variables]
                    raise KeyError(
                        f"""X-MET dataset does not contain variable "{var}". Present variables are: {", ".join(present_vars)}"""
                    )

        ds_xjsg = ds_xjsg.copy().swap_dims(
            {
                along_track_dim_xjsg: along_track_dim,
                across_track_dim_xjsg: across_track_dim,
            }
        )

        new_ds_msi = ds_msi.copy().swap_dims(
            {
                along_track_dim: f"{along_track_dim}_original",
                across_track_dim: f"{across_track_dim}_original",
            }
        )
        new_ds_msi[time_var] = ds_xjsg[time_var_xjsg].copy()

        lat_msi = ds_msi[lat_var].values.flatten()
        lon_msi = ds_msi[lon_var].values.flatten()
        coords_msi = sequence_geo_to_ecef(lat_msi, lon_msi)

        lat_jsg = ds_xjsg[lat_var_xjsg].values.flatten()
        lon_jsg = ds_xjsg[lon_var_xjsg].values.flatten()
        coords_jsg = sequence_geo_to_ecef(lat_jsg, lon_jsg)

        tree = cKDTree(coords_msi)
        dists, idxs = tree.query(coords_jsg, k=k)

        dims: str | tuple[str, str]
        for var in vars:
            if ds_msi[var].dims == (along_track_dim, across_track_dim):
                dims = (along_track_dim, across_track_dim)

                values = ds_msi[var].values
                values_flat = values.flatten()

                mask_nan = np.isnan(values_flat[idxs])

                _dists = dists
                _dists[mask_nan] = np.inf

                # Inverse distance weighting
                if k > 1:
                    weights = 1.0 / (_dists + eps)
                    weights /= np.sum(weights, axis=1, keepdims=True)
                else:
                    weights = np.ones(idxs.shape)

                if k > 1:
                    _v = values_flat[idxs]

                    if np.issubdtype(_v.dtype, np.floating):
                        m = np.all(np.isnan(_v), axis=1)
                        _v[np.isnan(_v)] = 0.0
                        _v[m] = np.nan

                    result = np.sum(_v * weights, axis=1)

                    new_values = result
                else:
                    new_values = values_flat[idxs]

                new_values = new_values.reshape(ds_xjsg.latitude_swath.shape)

                new_var = f"{var}"
                new_ds_msi[new_var] = (dims, new_values)
                new_ds_msi[new_var].attrs = ds_msi[var].attrs
            elif var not in _SKIP_VARS and var in ds_msi and var in ds_xjsg:
                new_ds_msi[var] = ds_xjsg[var].copy()
                new_ds_msi[var].attrs = ds_xjsg[var].attrs
            else:
                continue

        return new_ds_msi



import os
import re


def search_files_by_regex(root_dirpath: str, regex_pattern: str) -> list[str]:
    """
    Recursively searches for files in a directory that match a given regex pattern.

    Args:
        root_dirpath (str): The root directory to start the search from.
        regex_pattern (str): A regular expression pattern to match file names against.

    Return:
        list[str]: A list of absolute file paths that point to files with matching names.

    Raises:
        FileNotFoundError: If the root directory does not exist.
        re.error: If the given pattern is not a valid regular expression.
    """
    if not os.path.exists(root_dirpath):
        raise FileNotFoundError(
            f"{search_files_by_regex.__name__}() Root directory does not exist: {root_dirpath}"
        )

    filepaths = []
    for dirpath, _, filenames in os.walk(root_dirpath):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if re.search(regex_pattern, filename):
                filepaths.append(filepath)
    return filepaths



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    BSC_LABEL,
    DEPOL_LABEL,
    ELEVATION_VAR,
    EXT_LABEL,
    LR_LABEL,
    TROPOPAUSE_VAR,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_aice(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens ATL_ICE_2A file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="JSG_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        time_var="time",
        tropopause_var="tropopause_height",
        elevation_var="elevation",
    )

    var = "ice_water_content"
    ds[var].values = ds[var].values * 1e-3
    ds = rename_var_info(
        ds=ds,
        var=var,
        name="IWC",
        long_name="Ice water content",
        units="g/m$^3$",
    )
    var = "ice_water_content_error"
    ds[var].values = ds[var].values * 1e-3
    ds = rename_var_info(
        ds=ds,
        var=var,
        name="IWC error",
        long_name="Ice water content error",
        units="g/m$^3$",
    )
    ds = rename_var_info(
        ds=ds,
        var="ice_effective_radius",
        name="Ice effective radius",
        long_name="Ice effective radius",
        units="$\mu$m",
    )
    ds = rename_var_info(
        ds=ds,
        var="ice_effective_radius_error",
        name="Ice effective radius error",
        long_name="Ice effective radius error",
        units="$\mu$m",
    )

    return ds
